import pyFAI

class Eiger4m(pyFAI.detectors.Eiger):

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.aliases = ['Eiger4M']
        self.dim = self.shape = (2167, 2070)



import h5py
import numpy as np
import multiprocessing as mp
from ..XpcsAna.xpcsmethods import mat2evt
from ..misc.progressbar import progress
from . import EdfMethods as edf
from . import CbfMethods as cbf


def get_case(detector):
    """select the way data are read
    """
    if detector in ['id10_eiger_single_edf', 'pilatus_single_cbf', 'id02_eiger_single_edf']:
        case = 0
    elif detector in ['p10_eiger_h5', 'converted_h5', 'lambda_nxs']:
        case = 1
    elif detector == 'xcs_cspad_h5':
        case = 2
    elif detector == 'id02_eiger_multi_edf':
        case = 3
    else:
        case = -1
    return case

# def evt2prob( events, qroi, nbins ):

#     nq = len(events)
#     prob = np.empty((nq,nbins+1), np.float32)
#     for iq in range(nq):
#         counts = np.unique(events[iq], return_counts=1)[1]
#         tmp = np.histogram(counts, nbins-1, range=(1,nbins-1))[0]
#         npix = qroi[iq][0].size
#         prob[iq,:] = np.hstack((counts.sum(),npix-counts.size,tmp))/npix
#     return prob


def get_firstnlast(first, last, nf, dim):

    nf = nf - 1

    if last is None and first is None:
        last = (nf, *dim)
        first = (0,)*len(last)
    elif last is not None and first is None:
        if isinstance(last, (int, np.integer)):
            last = (last,)
        if len(last) == 1:
            last = (min([last[0],nf]), *dim)
        else:
            last = [min([last[0],nf]), *last[1:]]
        first = (0,)*len(last)
    elif last is None and first is not None:
        if isinstance(first, (int, np.integer)):
            first = (first,)
        if len(first) > 1:
            last = (nf, *dim)
        else:
            first = (first[0], 0, 0)
            last = (nf, *dim)
    elif last is not None and first is not None:
        if isinstance(first, (int, np.integer)):
            first = (first, 0, 0)
        else:
            first = (first[0], 0, 0)
        if isinstance(last, (int, np.integer)):
            last = (last, *dim)
        else:
            last = (last[0], *dim)
    elif last is not None and first is not None and len(last) == len(first) == 2:
        first = (0, first[1], first[0])
        last = (nf, last[1], last[0])
    return first, last


def alock(lock):
    """ acquire lock
    """
    if lock:
        lock.acquire()
    else:
        pass


def rlock(lock):
    """ release lock
    """
    if lock:
        lock.release()
    else:
        pass


def rearrange_tiles(arr, func):

    if arr.ndim > 3:
        for i in range(arr.shape[0]):
            if i == 0:
                tmp = func(arr[i])
                arranged_arr = np.empty((arr.shape[0], *tmp.shape))
                arranged_arr[i] = tmp
            else:
                arranged_arr[i] = func(arr[i])
        del arr
        arr = arranged_arr
    else:
        arr = func(arr)
    return arr


# -----------------------------------------------------------------------
# data format classes to provide specific functions to read data and get
# information on dataset shape
class dataset:
    """
    Class that provides all meta information, options, 
    correction functions for data loading
    """

    def __init__(self, opt):

        self.__dict__.update(opt)
        self.shape = None
        self.dstream = None
        self.variance = None
        self.chunk = None
        self.imgpf = None

    def update_shape(self, nimg, dim):

        self.shape = (nimg, *dim)

    def init_output(self, ):

        if self.method == 'full':
            self.dstream = np.empty(self.shape, dtype=self.dtype)
        elif self.method == 'average':
            pass
        elif self.method == 'queue_chunk':
            self.dstream = self.chunk
        elif self.method == 'events':
            if self.qroi is None:
                self.qroi = [self.mask]
                print('Qrois not defined. Using all pixels that are not masked.')
            self.dstream = [[] for i in range(len(self.qroi)+1)]
        else:
            pass

    def calc_mean(self, weighted=False):

        self.dstream = np.ma.masked_array(self.dstream)
        self.variance = np.ma.masked_array(self.variance)
        m, v = np.ma.average(self.dstream, axis=0, weights=1/self.variance,
                             returned=True)
        v = 1/v
        self.variance = v.data
        if weighted:
            self.dstream = m.data
        else:
            self.dstream = self.dstream.mean(0).data

    def process_chunk(self):

        arr = self.chunk.astype(self.dtype, copy=False)

        for correction in self.corrections:
            if correction == 'commonmode':
                for matr in arr:
                    for tile in matr:
                        tile -= self.commonmode(tile)

            elif correction == 'dark':
                arr -= self.dark[None, ...]

            # elif correction == 'mask':
            #     ind = (...,*mask)
            #     arr[ind] = mask_value

            elif correction == 'filter_value':
                arr[arr != self.filter_value] = 0

            elif correction == 'dropletize':
                arr = self.dropletize(arr, self.dropopt)

        if 'events' in self.method:
            # store mean intensity of frames in first list element
            if self.mask is not None:
                ind = (..., *self.mask)
                m = [arr[ind].mean(1),]
            else:
                m = [arr.mean((1, 2)),]

            # extend the dstream list by the intensities in the pixels
            # in the previously defined qrois
            for qi in range(len(self.qroi)):
                ind = (...,*self.qroi[qi])
                m.append(list(mat2evt(arr[ind])))

            arr = m

        self.chunk = arr


    def prepare_output(self):

        if self.output == 'original':
            pass
        else:
            if '2d' in self.output:
                if self.cond_rearrange_tiles():
                    self.dstream = rearrange_tiles(
                        self.dstream, self.arrange_tiles)
                    if self.variance is not None:
                        self.variance = rearrange_tiles(
                            self.variance, self.arrange_tiles)
                if 'sec' in self.output:
                    if self.cond_section():
                        qsec = self.qsec
                        self.dstream = self.dstream[..., qsec[0][0]:qsec[1][0]+1,
                                                    qsec[0][1]:qsec[1][1]+1]
                        if self.variance is not None:
                            self.variance = self.variance[..., qsec[0][0]:qsec[1][0]+1,
                                                          qsec[0][1]:qsec[1][1]+1]
                if self.dstream.shape[0] == 1:
                    self.dstream = np.squeeze(self.dstream)

    def cond_rearrange_tiles(self):

        if self.method == 'average':
            if self.dstream.ndim > 2:
                return True
        elif self.method == 'full':
            if self.dstream.ndim > 3:
                return True
        elif self.method == 'queue_chunk':
            if self.chunk.ndim > 3:
                return True
        else:
            return False

    def cond_section(self):

        qsec = self.qsec
        if qsec is not None:
            if self.dstream.shape[-2:] != (qsec[1][0]-qsec[0][0]+1, qsec[1][1]-qsec[0][1]+1):
                return True
            else:
                return False
        else:
            return False


class hdf5(dataset):

    def __init__(self, masterfile, opt, use_chunks=True):

        super().__init__(opt)
        self.masterfile = masterfile
        self.use_chunks = use_chunks

    def get_shape(self):

        alock(self.lock)

        with h5py.File(self.masterfile, 'r', driver=self.driver) as f:
            if self.extlinks:
                data_links = [f[self.datapath].name +
                              '/' + name for name in f[self.datapath]]
                datapath = data_links
            else:
                datapath = [self.datapath, ]
            dim = f[datapath[0]].shape
            nf = 0
            # determine number of images per file
            self.imgpf = dim[0]
            for d in datapath:
                nf += f[d].shape[0]

        rlock(self.lock)
        self.datapath = datapath
        self.shape = (nf, *dim[1:])
        return self.shape

    def load_chunk(self, indx=None):

        alock(self.lock)
        datapath = self.datapath[int(indx[0]//self.imgpf)]
        indx = (indx % self.imgpf).astype('int32')
        qsec = self.qsec
        with h5py.File(self.masterfile, 'r', driver=self.driver) as f:
            if indx is not None and qsec is not None:
                if len(f[datapath].shape) > 1:
                    arr = f[datapath][indx[0]:indx[-1]+1,
                                      qsec[0][0]:qsec[1][0]+1,
                                      qsec[0][1]:qsec[1][1]+1]
                else:
                    arr = f[datapath][indx[0]:indx[-1]+1]
            elif indx is not None:
                arr = f[datapath][indx[0]:indx[-1]+1]
            else:
                arr = f[datapath][...]

        rlock(self.lock)

        self.chunk = arr

    def start_reading_data(self):

        return None

    def stop_reading_data(self):

        return None


class sglimgfmt(dataset):

    def __init__(self, masterfile, datafiles, opt, use_chunks=True):

        super().__init__(opt)
        self.masterfile = masterfile
        if masterfile.endswith('edf') or masterfile.endswith('edf.gz'):
            self.get_image = edf.loadedf
        elif masterfile.endswith('cbf'):
            self.get_image = cbf.loadcbf
        else:
            raise KeyError('Data format neither edf nor cbf.')
        self.datafiles = datafiles
        self.use_chunks = use_chunks
        self.procs = []
        self.in_queue = []
        self.out_queue = []

    def get_shape(self):

        dim = self.get_image(self.masterfile).shape
        nf = len(self.datafiles)
        self.shape = (nf, *dim)
        return self.shape

    def load_chunk(self, indx):

        self.chunk = np.empty((len(indx), *self.shape[1:]), self.dtype)
        for ip in range(self.nprocs):
            self.in_queue[ip].put(
                self.datafiles[indx[0]+ip:indx[-1]+1:self.nprocs])

        ip = 0
        for i in range(len(indx)):
            self.chunk[i] = self.out_queue[ip].get()
            ip += 1
            if ip == self.nprocs:
                ip = 0

    def load_proc(self, in_queue, out_queue, qsec=None, dtype=None):

        while True:
            from_queue = in_queue.get()
            if from_queue is None:
                break
            for f in from_queue:
                matr = self.get_image(f).astype(dtype)
                if qsec is not None:
                    matr = matr[qsec[0][0]:qsec[1][0]+1, qsec[0][1]:qsec[1][1]+1]
                out_queue.put(matr)

    def start_reading_data(self):

        for ip in range(self.nprocs):
            self.in_queue.append(mp.Queue(self.nprocs))
            self.out_queue.append(mp.Queue(self.nprocs))
            self.procs.append(mp.Process(target=self.load_proc,
                                         args=(
                                             self.in_queue[ip], self.out_queue[ip],),
                                         kwargs={'qsec': self.qsec,
                                                 'dtype': self.dtype}))
            self.procs[ip].start()

    def stop_reading_data(self):

        for ip in range(self.nprocs):
            self.in_queue[ip].put(None)
            self.in_queue[ip].close()
            self.in_queue[ip].join_thread()
            self.out_queue[ip].close()
            self.out_queue[ip].join_thread()
            self.procs[ip].join()


class multiedf(dataset):

    def __init__(self, masterfile, datafiles, opt, use_chunks=True):

        super().__init__(opt)
        self.masterfile = masterfile
        self.get_image = edf.loadedf
        self.datafiles = datafiles
        self.use_chunks = use_chunks
        self.procs = []
        self.in_queue = []
        self.out_queue = []

    def get_shape(self):

        dim = self.get_image(self.masterfile).shape
        nf = len(self.datafiles)
        self.shape = (nf, *dim)
        return self.shape

    def load_chunk(self, indx):

        self.chunk = np.empty((len(indx), *self.shape[1:]), self.dtype)
        for ip in range(self.nprocs):
            self.in_queue[ip].put(
                self.datafiles[indx[0]+ip:indx[-1]+1:self.nprocs])

        ip = 0
        for i in range(len(indx)):
            self.chunk[i] = self.out_queue[ip].get()
            ip += 1
            if ip == self.nprocs:
                ip = 0

    def load_proc(self, in_queue, out_queue, qsec=None, dtype=None):

        while True:
            from_queue = in_queue.get()
            if from_queue is None:
                break
            for f in from_queue:
                matr = self.get_image(f).astype(dtype)
                if qsec is not None:
                    matr = matr[qsec[0][0]:qsec[1]
                                [0]+1, qsec[0][1]:qsec[1][1]+1]
                out_queue.put(matr)

    def start_reading_data(self):

        for ip in range(self.nprocs):
            self.in_queue.append(mp.Queue(self.nprocs))
            self.out_queue.append(mp.Queue(self.nprocs))
            self.procs.append(mp.Process(target=self.load_proc,
                                         args=(
                                             self.in_queue[ip], self.out_queue[ip],),
                                         kwargs={'qsec': self.qsec,
                                                 'dtype': self.dtype}))
            self.procs[ip].start()

    def stop_reading_data(self):

        for ip in range(self.nprocs):
            self.in_queue[ip].put(None)
            self.in_queue[ip].close()
            self.in_queue[ip].join_thread()
            self.out_queue[ip].close()
            self.out_queue[ip].join_thread()
            self.procs[ip].join()


# ----------------------
# _*_ MAIN FUNCTION _*_
# ----------------------
def read_data(datafiles, detector=None, last=None, first=None, step=[1, 1, 1], qroi=None,
              qsec=None, verbose=True, method='full', output='2dsection', chunk_size=256,
              dtype=np.float32, var_weight=False, nprocs=1, datapath="", driver='stdio',
              extlinks=False, filter_value=False, dropopt=None, dropmask=None, xdata=None,
              indxQ=None, dataQ=None, lock=False, dark=None, commonmode=True,
              dropletize=False, mask=False, mask_value=-1, **kwargs):


    # ---------------------------------------------
    # Nested Functions only invoked by read_data()
    # ---------------------------------------------
    def make_chunks():
        """
        if data should be read in chunks, this functions createas a list of
        chunked image indices
        """
        if verbose:
            print('Loading data in chunks of {} images.'.format(chunk_size))

        # old way
        # chunks = [np.arange(first[0] + i*chunk_size,
        #                     first[0] + min([min([(i + 1) * chunk_size, last[0]]), nimg]),
        #                     step[0])
        #           for i in range(np.ceil(nimg / chunk_size).astype(np.int32))]
        # new chunks
        ind_arange = np.arange(first[0],last[0]+1)
        bins = np.arange(0, nf, chunk_size)
        digitized = np.digitize(ind_arange, bins)
        chunks = [ind_arange[np.where(digitized==i)] for i in np.unique(digitized)]

        return chunks

    # --------------------
    # Beginning main code
    # --------------------

    # initialize reading options
    options = locals()
    options['corrections'] = []

    if dark is not None:
        if verbose:
            print('Doing dark subtraction.')
        options['dark'] = dark.astype(dtype)
        options['corrections'].append('dark')

    if isinstance(mask, np.ndarray):
        if qsec is not None and 'sec' in output:
            mask = mask[qsec[0][0]:qsec[1][0]+1, qsec[0][1]:qsec[1][1]+1]
        # options['corrections'].append('mask')

    if xdata is not None:
        try:
            h5opt = xdata.h5opt
        except:
            h5opt = {}
        if commonmode == 1 and 'commonmode' in h5opt:
            options['commonmode'] = h5opt['commonmode']
            options['corrections'].append('commonmode')
            if verbose:
                print('Using commonmode correction.')
        if dropletize == 1 and 'dropletize' in h5opt:
            options['dropletize'] = h5opt['dropletize']
            options['dropopt'] = dropopt
            options['corrections'].append('dropletize')
            if verbose:
                print('Images are dropletized.')
        if 'ExternalLinks' in h5opt:
            options['extlinks'] = h5opt['ExternalLinks']
            if options['extlinks'] and verbose:
                print('H5 file using external links.')
        if datapath == '' and 'data' in h5opt:
            options['datapath'] = h5opt['data']
        if 'chunk_size' in h5opt:
            chunk_size = h5opt['chunk_size']
        if 'arrange_tiles' in h5opt and '2d' in output:
            options['arrange_tiles'] = h5opt['arrange_tiles']
            if verbose:
                print('Rearranging tiles.')
        if 'mask' in vars(xdata.setup):
            mask = xdata.setup.mask.copy()
            if qsec is not None and 'sec' in output:
                mask = xdata.setup.qsec_mask

    if isinstance(mask, np.ndarray):
        mask = np.where(mask)
        if qsec is not None:
            mask = (mask[0]-qsec[0][0], mask[1]-qsec[0][1])
        options['mask'] = mask
    else:
        mask = None

    if qroi is not None and 'sec' in output:
        options['qroi'] = qroi
        for iq in range(len(qroi)):
            options['qroi'][iq] = (qroi[iq][0]-qsec[0][0], qroi[iq][1]-qsec[0][1])

    if 'sec' not in output:
        options['qsec'] = None
        qsec = None

    # dcls is the data class that contains all necessary functions
    # to deal with different file formats
    if detector is None and xdata is not None:
        detector = xdata.fmtstr

    case = get_case(detector)
    datafiles = np.array(datafiles)
    masterfile = datafiles[0]

    if case == 0:
        dcls = sglimgfmt(masterfile, datafiles, options)
    elif case in [1, 2]:
        dcls = hdf5(masterfile, options)
    elif case in [3]:
        dcls = multiedf
    else:
        raise ValueError('Case %d not defined.' % case)

    shape = dcls.get_shape()
    nf = shape[0]
    dim = shape[1:]

    if output == 'shape':
        return shape

    first, last = get_firstnlast(first, last, nf, dim)

    if qsec is not None and len(dim) < 3:
        dim = list(dim)
        dim[-2:] = (qsec[1][0]-qsec[0][0]+1, qsec[1][1]-qsec[0][1]+1)

    imgindx = np.arange(first[0], last[0]+1, step[0])
    nimg = len(imgindx)

    dcls.update_shape(nimg, dim)

    if verbose:
        print('First images is: ', first[0], flush=1)
        print('Last image is: ', last[0], flush=1)

    nargin = nimg
    if dcls.use_chunks:
        if chunk_size is None:
            chunk_size = nimg
        chunks = make_chunks()
        nargin = len(chunks)
        dcls.chunks = chunks

    dcls.init_output()

    # start loading data
    dcls.start_reading_data()

    if method == 'full':
        # process full data set
        for i in range(nargin):
            progress(i, max([nargin, 1]))

            dcls.load_chunk(chunks[i])
            dcls.process_chunk()
            dcls.dstream[chunks[i]-first[0]] = dcls.chunk

        dcls.prepare_output()
        progress(1, 1)

    elif method == 'events':
        # process full data set
        for i in range(nargin):
            progress(i, max([nargin, 1]))

            dcls.load_chunk(chunks[i])
            dcls.process_chunk()
            for qi in range(len(dcls.dstream)):
                if i == 0:
                    dcls.dstream = dcls.chunk.copy()
                    break
                else:
                    if qi == 0:
                        dcls.dstream[qi] = np.append(
                            dcls.dstream[qi], dcls.chunk[qi])
                    else:
                        for j in range(3):
                            if j == 1:
                                dcls.chunk[qi][j] += chunks[i][0] - first[0]
                            dcls.dstream[qi][j] = np.append(
                                dcls.dstream[qi][j], dcls.chunk[qi][j])

        progress(1, 1)

    elif method == 'average':
        # average data in time dimension
        for i in range(nargin):
            progress(i, max([nargin, 1]))

            dcls.load_chunk(chunks[i])
            dcls.process_chunk()
            if i == 0:
                dcls.dstream = np.empty(
                    (len(chunks), *dcls.chunk.shape[1:]), dtype=dtype)
                dcls.variance = dcls.dstream.copy()
            dcls.dstream[i] = dcls.chunk.mean(0)
            dcls.variance[i] = dcls.chunk.var(0)

        dcls.calc_mean(weighted=var_weight)
        dcls.prepare_output()
        progress(1, 1)

    elif method == 'queue_chunk':
        # pushing chunks to a queue for external analysis classes
        while not indxQ.empty():
            indx, chunk = indxQ.get()
            dcls.load_chunk(chunk)
            dcls.process_chunk()
            dcls.dstream = dcls.chunk
            dcls.prepare_output()
            dataQ.put((indx, dcls.dstream))

    # done reading data
    dcls.stop_reading_data()

    if method == 'queue_chunk':
        return None
    elif method == 'average':
        return dcls.dstream, dcls.variance/np.sqrt(nimg)
    else:
        return dcls.dstream



#! /usr/bin/env python
import numpy as np
from matplotlib import pyplot as plt
import fit.fitPoissonGamma as fpg


def plotpoissongamma(kb, p, kv, M, dkb=0, dp=0, dM=0, cmap='tab20', ind_var='kb', confint=None,
                     q=0, ax=None, xlim=None, axscale=None, **kwargs):
    if ax is None:
        ax = plt.gca()
    
    if ind_var == 'kb':
        #axis scale
        if axscale is None:
            xscale = 'log'
            yscale = 'log'
        else:
            xscale = axscale[0]
            yscale = axscale[1]
            
        # define colormap
        cmap = plt.get_cmap(name=cmap)
        ci = np.arange(kv.size)
        
        for i,k in enumerate(kv):
            kbi = kb.copy()
            pi = p[:,i]
            dpi = dp[:,i]
            c1 = cmap(ci[i])
            c2 = cmap(ci[i])
            ind = pi>0
            kbi = kbi[ind]
            dkbi = dkb.copy()[ind]
            pi = pi[ind]
            dpi = dpi[ind]
            # plot data 
            ax.errorbar(1/kbi, pi, yerr=dpi, xerr=1/kbi**2*dkbi, marker='o', linestyle='',
                           color=c2, markeredgewidth=.1,  markeredgecolor='gray',
                            label=r'$k={0:d}$'.format(int(k)))
            yl = ax.get_ylim()
            
            # define x-axis limits
            if xlim is None:
                kbs = kbi[(kbi>0)&(kbi<np.inf)]
                if kbs.size > 0:
                    xmin = np.min(1/kbs)*0.9
                    xmax = np.max(1/kbs)*1.1
                else:
                    xmin, xmax = ax.get_xlim()
            else:
                xmin, xmax = xlim

            # kb vector to plot fit
            kbf = np.logspace(np.log10(1/xmax),np.log10(1/xmin),50)

            # plot confident intervals
            if confint is not None:
                for Mc in confint:
                    pbf_lb = fpg.PoissonGamma(kbf, k=k, M=Mc)
                    ax.plot(1/kbf, pbf_lb, '--', color='gray')
                    
            if not np.isnan(M):
                # plot fit        
                # calculate estimated function and plot fit
                pbf = fpg.PoissonGamma(kbf, k=k, M=M)
                ax.plot(1/kbf, pbf, '-', color=c1)
                ax.annotate(r'$\beta={0:.2g}\pm{1:.2g}$'.format(1/M, 1/M**2*dM), xy=(.8,.1),
                            xycoords='axes fraction', ha="center", va="center",
                            bbox=dict(boxstyle="round", ec=(0.57, 0.57, 0.57), fc='w'))

        ax.set_yscale(yscale)
        ax.set_xscale(xscale)
        ax.set_ylim(yl)
        ax.set_xlabel(r'1/$\langle \mathrm{{k}}\rangle$')
        ax.set_ylabel('photon probability')
        ax.set_title(r'$\mathsf{{q}} = {:.2f}\times 10^{{-3}}\,\mathsf{{\AA}}^{{-1}}$'.format(q*1e3))
        ax.legend(loc='best')
        
    if ind_var == 'k' or ind_var == 'ks2':
        
        # define colormap
        cmap = plt.get_cmap(name=cmap)
        nplots = p.shape[0]
        ci = np.arange(nplots)
        
        for i in range(nplots):
            kbi = kb[i]
            pi = p[i,:]
            dpi = dp[i,:]
            c1 = 'gray'
            c2 = cmap(ci[i])
            
            # define x-axis limits
            if xlim is None:
                xmin = np.min(kv)
                xmax = np.max(kv)
            else:
                xmin, xmax = xlim

            # kb vector to plot fit
            kvf = np.linspace(xmin, xmax, 100)

            # plot confident intervals
            if confint is not None:
                for Mc in confint:
                    pbf_lb = fpg.PoissonGamma_indk(kvf, kb=kbi, M=Mc)
                    ax.plot(1/kvf, pbf_lb, '--', color='gray')

            if not np.isnan(M):
                # plot fit        
                # calculate estimated function and plot fit
                if (kvf.max()+M)>160:
                    pbf = np.exp(fpg.PoissonGamma_indk_approx(kvf, kb=kbi, M=M))
                else:
                    pbf = fpg.PoissonGamma_indk(kvf, kb=kbi, M=M)
                ax.plot(kvf, pbf, '-', color=c1)
                ax.annotate(r'$\beta={0:.2g}\pm{1:.2g}$'.format(1/M, 1/M**2*dM), xy=(.8,.1),
                            xycoords='axes fraction', ha="center", va="center",
                            bbox=dict(boxstyle="round", ec=(0.57, 0.57, 0.57), fc='w'))


            # plot data 
            ax.errorbar(kv, pi, yerr=dpi, marker='o', linestyle='',
                        color=c2, markeredgewidth=.1,  markeredgecolor='gray',
                        label=r'$\langle k\rangle={0:.3g}$'.format(kbi))

        if axscale is None:
            xscale = 'linear'
            yscale = 'log'
        else:
            xscale = axscale[0]
            yscale = axscale[1]

        ax.set_yscale(yscale)
        ax.set_xscale(xscale)

        ax.set_xlabel(r'k')
        ax.set_ylabel('photon probability')
        ax.set_title(r'$\mathsf{{q}} = {:.2f}\times 10^{{-3}}\,\mathsf{{\AA}}^{{-1}}$'.format(q*1e3))
        ax.legend(loc='best')



from .pysaxs3 import pysaxs
from .defineqrois import defineqrois
from .find_center import find_center

class Saxs:

    def __init__(self, ):

        pass


    def pysaxs(*args, **kwargs):
        return pysaxs(*args, **kwargs)

    def defineqrois(*args, **kwargs):
        defineqrois(*args, **kwargs)

    def find_center(*args, **kwargs):
        return find_center(*args, **kwargs)




#! /usr/bin/env python
import numpy as np
from misc.imgkicker import imgkicker
from scipy.stats import norm


def find_kb(sel, dsel, kbr):
    if (len(kbr) == 1) or (kbr[0] == kbr[1]):
        ind = np.append(0,np.argmin(np.abs(sel[1:,0]-kbr[0]))+1)
    else:
        ind = np.append(0,np.where((sel[1:,0]>kbr[0])&(sel[1:,0]<kbr[1]))[0]+1)
    if ind.size == 1:
        raise ValueError('No data found in kb range.')
    sel = sel[ind,:]
    dsel = dsel[ind,:]
    return sel, dsel

def findk(kr, k_range):
    iin = np.array([np.where(kr==ki)[0][0] if ki in kr else -1 for ki in k_range])
    iout = np.append(0,np.where(iin>=0)[0]+1)
    iin = np.append(0,np.delete(iin,np.where(iin<0))+1)
    if iin.size == iout.size == 1:
        raise ValueError('Could not find any k values defined in k_range. Mean counts: {:.2g}'.format(np.mean(kr)))
    return iin, iout

def get_all(sel, d1, ind, xsvs, dxsvs, roi_idx, k_range, indk, kbr):
    #init output array
    sel = np.vstack((sel,np.zeros((np.sum(d1),sel.size))))
    dsel = sel.copy()
    for ii,i in enumerate(ind):
        kr = xsvs[i][roi_idx,0,1:]
        iin, iout = findk(kr, k_range)
        dind = np.arange(1+np.sum(d1[:ii]),1+np.sum(d1[:ii+1]))
        sel[dind[:,None],iout] = xsvs[i][roi_idx,indk[ii][:,None],iin]
        dsel[dind[:,None],iout] = dxsvs[i][roi_idx,indk[ii][:,None],iin]
    sel, dsel = find_kb(sel, dsel, kbr)
    return sel, dsel

def selectdata(xsvs, t, k_range=None, t_int=None, dxsvs=None, method='all', roi_idx=0, t_short=0.,
               t_long=np.inf, t_readout=0., att=0, bw=1., series=None, kicker='imgkicker', kbr=(0,np.inf), pbthres=None, verbose=False):
    '''method to select photon probablity data
    '''
    
    # define k-range
    if k_range is None:
        k_range = np.arange(20)

    # first column of output array contains k values
    sel = np.append(0,k_range)
    
    # define integration time vector if not passed
    if t_int is None:
        t_int = np.arange(len(xsvs))
    
    if dxsvs is None:
        dxsvs = np.ones_like(xsvs).astype(np.uint8)

    ind = np.where(t_int==t)[0]
    if ind.size == 0:
        raise ValueError('Exposure time {:.2g} was not found.'.format(t))
    if verbose:
        print('Found {} series: {}.'.format(len(ind), ind))
    if series is not None:
        ind = ind[series]
    
    # use imagekicker to select only a certain time region from a data set
    # the image indices of each series are saved in the list indk
    # d1 contains the number of 1s (or selected images) in indk
    d1 = []
    indk = []
    for ii,i in enumerate(ind):
        if type(kicker) == str:
            if kicker == 'imgkicker':
                tmp = imgkicker(t, xsvs[i].shape[1], t_short, t_long, t_readout, att)
                tmp[:2] = 0
                indk.append(np.where(tmp)[0])
        else:
            indk.append(kicker+1)
        d1.append(indk[ii].size)
    d1 = np.array(d1)    
    
    ###---MAIN PART--- select data based on four different methods:
    ###--------------- all -- bin_all -- bin_series -- bin
    # get probability data for from all images
    if method == 'all':
        sel, dsel = get_all(sel, d1, ind, xsvs, dxsvs, roi_idx, k_range, indk, kbr)

    # average over entire series by fitting with a Gaussian distribution
    elif method == 'bin_series':
        #init output array
        sel = np.vstack((sel,np.zeros((ind.size,sel.size))))
        dsel = sel.copy()
        for ii,i in enumerate(ind):
            tmp = np.zeros((d1[ii],sel.shape[-1]))
            tmperr = tmp.copy()
            kr = xsvs[i][roi_idx,0,1:]
            iin, iout = findk(kr, k_range)
            tmp[:,iout] = xsvs[i][roi_idx,indk[ii][:,None],iin]
            tmperr[:,iout] = dxsvs[i][roi_idx,indk[ii][:,None],iin]
            tmp, tmperr = find_kb(tmp, tmperr, kbr)
            for r in range(np.shape(tmp)[1]):
                col = tmp[1:,r]
                gz = np.isfinite(col)#np.where(col>0)[0]
                mu, std = (np.nanmean(col[gz]),np.nanstd(col[gz]))
                sel[ii+1,r] = mu
                dsel[ii+1,r] = std

    # average over every single data point
    elif method == 'bin_all':
        sel, dsel = get_all(sel, d1, ind, xsvs, dxsvs, roi_idx, k_range, indk, kbr)
        dsel[1,:] = np.nanstd(dsel[1:,:], axis=0)
        sel[1,:] = np.nanmean(sel[1:,:], axis=0)
        dsel = dsel[:2,:]
        sel = sel[:2,:]
        

    # more advanced binning
    elif method == 'bin':
        selt, dselt = get_all(sel, d1, ind, xsvs, dxsvs, roi_idx, k_range, indk, kbr)
        kv = selt[0,:]
        
        #define bins
        if bw < 1:
            if bw < 0:
                bw *= -1
                logscale = 1
            else:
                logscale = 0
            kb = sel[1:,0]
            kbrange = np.max(kb[kb<np.inf])-np.min(kb[kb>0])
            bw *= kbrange
            nb = np.ceil(kbrange/bw).astype(np.int32)
            
            if nb >= kb.size:
                raise ValueError('More bins than data points')
            
            bins = np.linspace(np.min(kb[kb>0]),np.max(kb[kb<np.inf]),nb+1)   
            if logscale:
                bins = np.logspace(np.log10(bins[-1]),np.log10(bins[0]),nb+1)
                bins = bins[::-1]

            tmp = sel.copy()
            sel = np.zeros((nb,kv.size))
            dsel = sel.copy()
            #calculated mean and standard deviation of histograms
            for ii in range(bins.size-1):
                idx = (kb>=bins[ii]) & (kb<bins[ii+1])
                ttmp = tmp[idx,:]
                for s in range(kv.size):
                    mu, std = (np.nanmean(ttmp[:,s]), np.nanstd(ttmp[:,s]))
                    if np.isnan(mu):
                        print(ttmp[:,s])

                    sel[ii,s] = mu
                    dsel[ii,s] = std
                    
        elif bw >= 1:
            
            for ii in range(len(ind)):
                if ii == 0:
                    sels = selt[1:d1[ii],:]
                    dsels = dselt[1:d1[ii],:]
                else:
                    sels = selt[1+d1[ii-1]:1+d1[ii],:]
                    dsels = dselt[1+d1[ii-1]:1+d1[ii],:]
                r = int(np.shape(sels)[0] % bw)
                if r > 0:
                    sels = sels[:-r]
                    dsels = dsels[:-r]
                nkb = int(np.shape(sels)[0]/bw)
                nk = sels.shape[1]
                sels = np.mean(np.reshape(sels,(bw,nkb,nk),'F'),axis=0)
                dsels = np.mean(np.reshape(dsels,(bw,nkb,nk),'F'),axis=0)
                if ii == 0:
                    sel = np.vstack((selt[0,:],sels))
                    dsel = np.vstack((dselt[0,:],dsels))
                else:
                    sel = np.vstack((sel,sels))
                    dsel = np.vstack((dsel,dsels))
        
    kv = sel[0,1:]
    kb = sel[1:,0]
    dkb = dsel[1:,0]
    pba = sel[1:,1:]
    dpba = dsel[1:,1:]   
    
    if pbthres is not None:
        # print('Length of pba is {}'.format(len(pba)))
        # print(pba)
        for i in range(len(pba)):
            if i == 0:
                ind = (pba[i]>pbthres).astype(np.bool)
            else:
                ind = ind & (pba[i]>pbthres).astype(np.bool)

        if all(~ind):
            raise ValueError('All pba below threshold.')
        ind = np.where(ind)[0]
        kv = kv[ind]
        pba = pba[:,ind]
        dpba = dpba[:,ind]

    # return selected data
    return kv, kb, dkb, pba, dpba



import readline

def histsave(f, histfile = "./.save_history"):
    readline.clear_history()
    try:
        readline.read_history_file(histfile)
        h_len = readline.get_current_history_length()
    except FileNotFoundError:
        open(histfile, 'wb').close()
        h_len = 0
        
    readline.add_history(f)
    readline.write_history_file(histfile)
    
def get_history_list(histfile="~/xana_save_history", ftype='setup'):
    histlist = []
    with open(histfile, 'r') as f:
        line = f.readline()
        while line:
            if ftype in line.split('/')[-1]:
                histlist.append(line.rstrip('\n'))
            line = f.readline()
    return np.unique(histlist)



import numpy as np
import lmfit
import re
from fit.BandyFunctions import *


def fitbandy( et, M, err=None, mode='M', scaling='logxlogy', start=None, lb=None, ub=None, fix=None, doplot=0, h_plot=None,
                ax=None, output='pars', xl=None, ylim=None, cond=None, color=None, linestyle='-'):
    pars = lmfit.Parameters()
    pars.add('t', value=start[0], min=lb[0], max=ub[0], vary=1)
    pars.add('g', value=start[1], min=lb[1], max=ub[1], vary=1)
    pars.add('b', value=start[2], min=lb[2], max=ub[2], vary=1)
    pars.add('a', value=0, vary=0)
            
    if 'off' in mode:
        if len(start) == 3:
            pars['a'].set(vary=1)
        elif len(start) == 4:
            pars['a'].set(value=start[3], min=lb[3], max=ub[3], vary=1)
            
    if fix is not None:
        for vn in fix.keys():
            pars[vn].set(value=fix[vn], vary=0)

    if err is not None:
        if 'logy' in scaling:
            wgt = (M*np.log(10)/err)**2 
        else:
            wgt = 1./err**2
    else:
        err = np.ones_like(M)
        wgt = np.ones_like(M)
        
    Mp = M
    etp = et
    errp = err
    if cond is not None:
        et = et[cond]
        M = M[cond]
        err = err[cond]
        wgt = wgt[cond]
    
    if 'logy' in scaling:
        M = np.log10(M)
    if 'logx' in scaling:
        wgt *= et**10
     
    if 'M' in mode:
        func = bandy_logMa
    mod  = lmfit.Model(func)
    out = mod.fit(M, pars, x=et, weights=wgt)
    
    if doplot:
        if xl is None:
            if ax is None:
                ax = plt.gca()
            xl = ax.get_xlim()
            if xl[0] == 0:
                xl = (np.min(et)*0.9,np.max(et)*1.1)
        xf = np.logspace(np.log10(xl[0]),np.log10(xl[1]),50)
        bandyf = 1/bandy_V2a(xf, g=out.best_values['g'], t=out.best_values['t'], b=out.best_values['b'], a=out.best_values['a'])
        
        if 'V2' in doplot:
            bandyf = 1/bandyf
            Mp = 1/Mp
            errp = Mp**2*errp
        if 'g2' in doplot:
            Mp += 1.
            bandyf += 1.
        
        if 'legend' in doplot:
            pard = {'t':r'$t_0: {:.2g}\mathrm{{s}},\,$',
                    'g':r'$\gamma: {:.2g},\,$',
                    'b':r'$\mathrm{{b}}: {:.2g},\,$',
                    'a':r'$\mathrm{{a}}: {:.2g},\,$'}
            labstr = ''
            for vn in 'tgba':
                if vn in out.var_names:
                    labstr += pard[vn].format(out.best_values[vn])
                elif fix is not None and vn in fix.keys():
                    labstr += 'fix '+ pard[vn].format(fix[vn])
                else:
                    labstr += pard[vn].format(0)
        else:
            labstr = ''
        
        pl = []
        pl.append(ax.plot(xf, bandyf, linestyle, label=labstr, color=color, linewidth=1))
        if 'data' in doplot:
            pl.append(ax.errorbar( etp, Mp, yerr=errp, marker='d', markersize=3.5, linestyle=''))
        
        if color is None:
            if h_plot is not None:
                color = h_plot[0].get_color()
            elif 'data' in doplot:
                color = pl[0][0].get_color()
            else:
                color = 'gray'
        for p in pl:
            p[0].set_color(color)
            
        ax.legend()
        if 'report' in doplot:
            print(out.fit_report())
            
    if output == 'pars':
        pars =  np.zeros((4,2))
        for i,vn in enumerate(out.var_names):
            pars[i,0] = out.best_values[vn]
            param = list(out.params.values())
            pars[i,1] = param[i].stderr
        return pars
    elif output == 'fit':
        return out

def fitBandy2(et, M, dM=None, mode='logM', start=None, lb=None, ub=None):
    def residual(pars, x, data=None, eps=None):
        # unpack parameters:
        #  extract .value attribute for each parameter
        parvals = pars.valuesdict()
        period = parvals['period']
        shift = parvals['shift']
        decay = parvals['decay']

        if abs(shift) > pi/2:
            shift = shift - sign(shift)*pi

        if abs(period) < 1.e-10:
            period = sign(period)*1.e-10

        model = parvals['amp'] * sin(shift + x/period) * exp(-x*x*decay*decay)

        if data is None:
            return model
        if eps is None:
            return (model - data)
        return (model - data)/eps
    pars = lmfit.Parameters()
    pars.add('t', value=start[0], min=lb[0], max=ub[0], vary=1)
    pars.add('g', value=start[1], min=lb[1], max=ub[1], vary=1)
    if dM is not None:
        wgt = 1./dM**2
    else:
        wgt = np.ones_like(M)
    if mode == 'M':
        mod = lmfit.Model(bandy_M)
    elif mode == 'logM':
        mod = lmfit.Model(bandy_logM)
    elif mode == 'logMa' or mode == 'logMa_off':
        mod  = lmfit.Model(bandy_logMa)
        if 'off' in mode:
            pars.add( 'a', value=0, min=0, vary=1)
        else:
            pars.add( 'a', value=0, vary=0)
        pars.add( 'b', value=start[2], min=lb[2], max=ub[2], vary=1)
    elif mode == 'logV2a':
        mod  = lmfit.Model(bandy_V2a)
        wgt = np.log10(wgt)
        pars.add( 'a', value=0, vary=0)
        pars.add( 'b', value=start[2], min=lb[2], max=ub[2], vary=1)
    if 'log' in mode:
        M = np.log10(M)
        if dM is not None:
            wgt = (1./np.log10(dM)**2)
    return mod.fit(M, pars, x=et, weights=wgt)
    




from time import time
import numpy as np
import pickle as pkl
from multiprocessing import Process, Queue
from .mp_prob import mp_prob
from ..misc.progressbar import progress
import sys


def pyxsvs( data, qroi, nbins=15, t_e=1., qv=None, method='full', nprocs=1,
            verbose=1, qsec=(0,0)):
    """Calculate photon proababilities.
    """
    time0 = time()
    lqv = len(qroi)
    rlqv = range(lqv)

    if qv is None:
        qv = np.arange(lqv)

    if isinstance(data, np.ndarray):
        nf, dim2, dim1 = np.shape(data)
        def get_chunk():
            return data
    elif isinstance(data, dict):
        nf = data['nimages']
        def get_chunk():
            return data['dataQ'].get()

    if verbose:
        print('Number of images is:', nf)
        print('Loading data in chunks.')
        print('Number of ROIs: ', lqv)

    lind = []
    total_pixels = 0
    for iq in rlqv:
        npixel = len(qroi[iq][0])
        lind.append(npixel)
        total_pixels += npixel

    nprocs = min(nprocs,lqv) # cannot use more processes than q-values
    tmp_pix = 0
    if nprocs >= lqv:
        q_sec = np.arange(lqv+1)
    else:
        q_sec = [lqv,]
        for iq in rlqv[::-1]:
            tmp_pix += lind[iq]
            if tmp_pix >= np.floor(total_pixels/nprocs):
                q_sec.append(iq)
                total_pixels -= tmp_pix
                nprocs -= 1
                tmp_pix =  0
            if iq == 0 or nprocs == 0:
                q_sec.append(0)
    q_sec = np.unique(q_sec)
    del tmp_pix
    nprocs = len(q_sec) - 1
    print('Using {} processes.'.format(nprocs))

    trace = np.empty((nf,lqv))

    #----multiprocessing----
    # create lists of queues and processes
    qur = []
    qure = []
    pcorr = []
    for i in range(nprocs):
        qur.append(Queue(16))
        qure.append(Queue(1))
    for i in range(nprocs):
        q_beg = q_sec[i]
        q_end = q_sec[i+1]
        pcorr.append(Process(target=mp_prob, args=(method, nbins, nf,
                lind[q_beg:q_end], q_end-q_beg, qur[i], qure[i])))

    # start processes
    for i in range(nprocs):
        pcorr[i].start()
    #-----------------------

    # start processing data
    tcalc_cum = 0
    t0 = 0
    last_chunk = -1
    while t0 < nf:
        progress(t0,nf)

        c_idx, chunk = get_chunk()
        chunk_size = chunk.shape[0]
        idx = slice(t0,t0+chunk_size)

        chunk_diff = c_idx - last_chunk
        if chunk_diff != 1:
            raise IOError('Chunks have been read in wrong order: chunk index difference is % and not 1.' %
                             chunk_diff)
        last_chunk = c_idx

        for jj,(i,j) in enumerate(zip(q_sec[:-1], q_sec[1:])):
            tmp_put = []
            for qi in range(i,j):
                q0 = qroi[qi][0] - qsec[0]
                q1 = qroi[qi][1] - qsec[1]
                roi = chunk[:,q0,q1]
                trace[idx,qi] = roi.mean(-1)
                tmp_put.append(roi)
            qur[jj].put(tmp_put)

        t0 += chunk_size

    progress(1,1)

    # read data from output queue and shut down processes
    from_proc = []
    for i in range(nprocs):
        from_proc.append(qure[i].get())
        pcorr[i].join()
        qure[i].close()
        qure[i].join_thread()

    # concatenate data from different processes
    p = from_proc[0][0]
    tcalc_cum = from_proc[0][1]
    for i in range(1,nprocs):
        p = np.concatenate((p,from_proc[i][0]), axis=0)
        tcalc_cum = max(tcalc_cum,from_proc[i][1])

    # initialize correlation array 'cc'
    prob = np.zeros((lqv+1,nbins+2,nf), dtype=np.float32)
    prob[1:,1:] = p
    prob[0,0,0] = t_e
    prob[1:,0,0] = qv
    prob[0,1:,0] = np.append(0,np.arange(nbins))

    if verbose:
        print("\rFinished calculating correlation functions.") 

    del p, from_proc

    if verbose:
        print('Elapsed time: {:.2f} min'.format((time()-time0)/60.))
        print('Elapsed time for calulating probabilities: {:.2f} min'.format(tcalc_cum/60.))

    probd = {'prob':prob, 't_exposure':t_e, 'trace':trace, 'qv':qv, 'qroi':qroi,}

    return probd




#! /usr/bin/env python
import numpy as np
from matplotlib import pyplot as plt
from ..Xfit.PoissonGammaDistribution import PoissonGamma as pg


def plot_poissongamma(x, y, p, M, cmap='tab20', ind_var='kb', confint=None,
                     q=None, ax=None, xlim=None, log='xy', **kwargs):
    if ax is None:
        ax = plt.gca()
    
    for i,pi in enumerate(p):
        print(pi, M)
        # plot data 
        ax.errorbar(1/x, y[i], yerr=None, marker='o', linestyle='',
                        label=ind_var+r'{0:d}'.format(int(pi)))

        # kb vector to plot fit
        xf = np.logspace(np.log10(1/x.max()),np.log10(1/x.min()),100)

        # plot confident intervals
        if confint is not None:
            for Mc in confint:
                pbf = pg(xf, Mc, pi, ind_var)
                ax.plot(1/xf, pbf, '--', color='gray')

        
        pbf = pg(xf, M[0][0], pi, ind_var)
        print(pbf)
        ax.plot(1/xf, pbf, '-')
        ax.annotate(r'$\beta={0:.2g}\pm{1:.2g}$'.format(1/M[0][0], 1/M[0][0]**2*M[0][1]), xy=(.8,.1),
                    xycoords='axes fraction', ha="center", va="center",
                    bbox=dict(boxstyle="round", ec=(0.57, 0.57, 0.57), fc='w'))

    if 'x' in log:
        ax.set_xscale('log')
    if 'y' in log:
        ax.set_yscale('log')
    ax.set_xlabel(r'1/$\langle \mathrm{{k}}\rangle$')
    ax.set_ylabel('photon probability')
#    ax.set_title(r'$\mathsf{{q}} = {:.2f}\times 10^{{-3}}\,\mathsf{{\AA}}^{{-1}}$'.format(q*1e3))
    ax.legend(loc='best')



from . import detectors
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
import os
import numpy as np


class Setup:
    """Xana Setup class containing the experimental parameters.
    """
    def __init__(self, **kwargs):
        """Init function of the Xana.Setup class.

        The class is initialized when creating a Xana object and contains all
        parameters or attributes describing the experimental setup.
        """
        self.detector = kwargs.pop('detector', 'eiger500k') #: detector used
        self.maskfile = kwargs.pop('maskfile', None)
        """str: path to the maskfile"""
        self.mask = self.load_mask() #: np.ndarray: mask array
        self.wavelength = None #: float: X-ray wavelength in Angstrom
        self.distance = None
        self.center = None
        self.ai = None
        self.qv_init = None
        self.phiv_init = None
        self.dqv = None
        self.qroi = None
        self.gproi = None
        self.qsec = None
        self.qsec_center = None
        self.qsec_mask = None
        self.qsec_dim = None
        self.qsec_ai = None
        self.qv = None
        self.phiv = None
        self.radii = None

    @property
    def detector(self):
        return self.__detector

    @detector.setter
    def detector(self, name):
        if name is None:
            self.__detector = None
            print('No sepcific detector defined.')
        else:
            self.__detector = detectors.grab(name)

    def __getstate__(self):
        d = dict(vars(self))
        d['detector'] = self.detector.aliases[0].lower()
        del d['ai'], d['qsec_ai'], d['_Setup__detector']
        return d

    def __setstate__(self, d):
        if 'lambda' in d:  # handle old syntax
            d['wavelength'] = d.pop('lambda')
        if 'ctr' in d and 'center'not in d:
            d['center'] = d['ctr']

        self.detector = d.pop('detector', None)
        self.__dict__.update(d)
        self.ai = self.update_ai()
        if d['qsec'] is not None:
            self.qsec_ai = self.update_ai(self.qsec_center)
        else:
            self.qsec_ai = None

    def make(self, **kwargs):
        keys = ['center', 'distance', 'wavelength', ]
        mesg = [('beam center x [pixel]', 'beam center y [pixel]'),
                ('sample-detector distance [m]',),
                ('wavelength [A]',),
                ]
        setup = {}
        for k, m in zip(keys, mesg):
            if k in kwargs:
                inp = kwargs[k]
            else:
                inp = []
                for mi in m:
                    inp.append(input(mi + '\t'))
            if k == 'center':
                inp = np.array(inp).astype('int32')
            else:
                inp = np.array(inp).astype('float32')
            setup[k] = inp
        self.__dict__.update(setup)
        self.ai = self.update_ai()

    def update_ai(self, center=None, nbins=1000):

        if center is None:
            center = self.center

        ai = AzimuthalIntegrator(detector=self.detector, dist=self.distance)
        ai.setFit2D(self.distance * 1000, center[0], center[1])
        ai.wavelength = self.wavelength * 1e-10
        return ai

    def load_mask(self):
        if isinstance(self.maskfile, str):
            self.maskfile = os.path.abspath(self.maskfile)
            # if self.maskfile.endswith('edf'):
            #     mask = loadedf(self.maskfile)
            if self.maskfile.endswith('npy'):
                mask = np.load(self.maskfile)
            else:
                print('Mask file not found. Continuing with no pixel masked.')
                mask = np.ones(self.detector.shape, dtype='bool')
        else:
            mask = np.ones(self.detector.shape, dtype='bool')
        return mask

    def from_dict(self):
        pass




import os
import pickle
from .Analysis import Analysis
from .Setup import Setup
from .Xdb import Xdb
from .misc.xsave import mksavdir, save_result, make_filename


class Xana(Xdb, Analysis):
    """Xana class to perform XPCS, XSVS and SAXS data analysis.
    """

    def __init__(self, sample='', savdir='./', setupfile=None, **kwargs):
        """__init__ of Xana.

        Args:
            sample (str): sample name (optional).
            **kwargs: kwargs passed to initialize the database (Xdb), the
                setup (Setup) and the Analysis class (Analysis).
        """

        self.savdir = savdir #: str: directory to save results in.
        self.sample = sample #: str: sample name appearing in the database.
        self.setupfile = setupfile
        """str: setupfile to load at the beginning, can also be loaded with loadsetup method."""

        Xdb.__init__(self, **kwargs)
        if self.setupfile is None:
            self.setup = Setup(**kwargs)
        else:
            self.loadsetup(self.setupfile)
        Analysis.__init__(self, **kwargs)


    def mksavdir(self, savdir, *args, **kwargs):
        """Create directory for saving results.

        Args:
            savdir (str): name of the directory for saving the results.
            **kwargs: kwargs can be `savhome` and `handle_existing`. The former is the
                directory in which savdir will be created. The latter sets the default
                behavior to use existing directories, raise and error or overwrite
                files.
        """

        self.savdir = mksavdir(savdir, *args, **kwargs)
        self.setup.savdir = self.savdir
        self.dbfile = self.savdir+'Analysis_db.pkl'
        self.load_db(handle_existing='overwrite')

    def loadsetup(self, filename=None):

        if filename is None:
            print('No setup defined.')
        else:
            path, fname = make_filename(self, 'setupfile', filename)
            fname = path + fname
            if os.path.isfile(fname):
                self.setupfile = fname
                self.setup = pickle.load(open(self.setupfile, 'rb'))
                print('Loaded setupfile:\n\t{}.'.format(self.setupfile))
            else:
                self.setupfile = None
                print('No setup defined.')

    def savesetup(self, filename=None, **kwargs):
        if filename is None:
            filename = input('Enter filename for setupfile.\t')
#        savd = copy.deepcopy(vars(self))
        self.setupfile = save_result(self.setup, 'setup', self.savdir, filename, **kwargs)



kb = 1.381e-23
R = 11e-9
T = 25
e_rheo = 10

def getD(eta, err=0):
    dD = 0
    D = kb*(T+273.15)/(6*np.pi*R*eta)
    if err:
        dD = D*err/eta        
    return D, dD

def geteta(D, err=0):
    deta = 0
    eta = kb*(T+273.15)/(6*np.pi*R*D*1e-18)
    if err:
        deta = eta*err/D
    return eta, deta

fig, ax = plt.subplots(2,2,figsize=(9,8))
ax = ax.ravel()
cmap = plt.get_cmap('Set1')

qp = np.arange(nq)

qvn = qv[qp]*10
qvn2 = qvn**2
x = np.linspace(np.min(qvn),np.max(qvn),100)
x2 = x*x

#-------plot decay rates--------
for i in range(2):
    y = 1/rates[qp,1+3*i,0]
    dy = y**2*rates[qp,1+3*i,1]
    ax[0].errorbar(qvn2[qp],y,dy,fmt='o',color=cmap(i), alpha=.6, label=r'$\Gamma_{}$'.format(i))
    
    nfl = [np.arange(9)]
    if i==0:
        nfl.append(np.arange(11,16))
        qp = np.arange(9)
        
    if i==1:
        continue
        
    for nf in nfl:
        nf = np.delete(nf,np.where(rates[nf,1,1]==0))
        D_mc, b_mc = emceefit_sl(qvn2[nf], y[nf], dy[nf])[:2]
        D_mc, b_mc = map(lambda x: (x[0],np.mean(x[1:])),[D_mc,b_mc])
        popt, pcov = np.polyfit(qvn2[nf], y[nf], 1, w=1/dy[nf], cov=1)
        perr = np.sqrt(np.diag(pcov))
        D_exp, b_exp = (popt[0],perr[0]), (popt[1],perr[1])

        y_dif = y[nf] / qvn2[nf]
        dy_dif = dy[nf] / qvn2[nf]
        D_dif = (np.sum(y_dif/dy_dif**2)/np.sum(1/dy_dif**2), np.sqrt(1/np.sum(1/dy_dif**2)))

        e_dif = geteta(*D_dif)
        e_exp = geteta(*D_exp)
        e_mle = geteta(*D_mc)
        D_rheo = getD(e_rheo)

        ax[0].plot(x2, np.polyval(popt,x2), color=cmap(i))
        #ax[0].plot(x2, np.polyval([D_dif[0],0],x2), color='gray')
        print('\nResults for {} decay:'.format(i+1))
        print('-'*20)
        print('D_rheo = {:.2f} nm2s-1'.format(D_rheo[0]*1e18))
        print(r'D_exp = {:.2f} +/- {:.2f} nm2s-1'.format(*D_exp))
        print(r'b_exp = {:.2f} +/- {:.2f} s-1'.format(*b_exp))
        print(r'D_mle = {:.4f} +/- {:.4f} nm2s-1'.format(*D_mc))
        print(r'b_mle = {:.4f} +/- {:.4f} s-1'.format(*b_mc))
        print(r'D_dif = {:.2f} +/- {:.2f} nm2s-1'.format(*D_dif))
        print('\neta_rheo = {:.2f} Pas'.format(e_rheo))
        print(r'eta_exp = {:.2f} +/- {:.2f} Pas'.format(*e_exp))
        print(r'eta_mle = {:.4f} +/- {:.4f} Pas'.format(*e_mle))
        print(r'eta_dif = {:.2f} +/- {:.2f} Pas'.format(*e_dif))

#-------plot KWW exponent--------
qp = np.arange(nq)
g = rates[qp,2,0]
dg = rates[qp,2,1]
ax[2].errorbar(qvn,g,dg,fmt='o',color=cmap(0), alpha=.6, label=r'$\gamma_{}$'.format(0))
qp = np.arange(9)
g = rates[qp,5,0]
dg = rates[qp,5,1]
ax[2].errorbar(qvn[qp],g,dg,fmt='s',color=cmap(1), alpha=.6, label=r'$\gamma_{}$'.format(1))
#-------plot nonergodicity parameter--------
def blc(q,L,k,lc):
    def A(q):
        return 4*np.pi/lc*q/k*np.sqrt(1-q**2/(4*k**2))
    return 2*(A(q)*L-1+np.exp(-A(q)*L))/(A(q)*L)**2

b = np.sum(rates[:,3::3,0],1)
db = np.sum(rates[:,3::3,1],1)

ax[1].errorbar(qvn2, b, db, fmt='o', color=cmap(0), alpha=.6, label=r'$f_0(q)$')

nf = np.arange(8)
nf = np.delete(nf,np.where(rates[nf,1,1]==0))
y = np.log(b[nf])
dy = db[nf]/b[nf]

popt1, cov1 = np.polyfit(qvn2[nf], y, 1, w=1/dy, cov=1)
cov1 = np.sqrt(np.diag(cov1))[0]
ax[1].plot(x2, exp(np.polyval(popt1,x2)), '-', color=cmap(0))
        #label=r'${:.3g}\, \exp{{(-q^2\cdot{:.3g}nm^{{2}})}}$'.format(np.exp(popt1[1]),-1*popt1[0]))
    
#----
b = rates[qp,6,0]
db = rates[qp,6,1]
ax[1].errorbar(qvn2[qp], b, db, fmt='s', color=cmap(1), alpha=.6, label=r'$f_1(q)$')
 
#----    

y_th = exp(popt[1])*blc(x, 1e-5, 2*np.pi/(12.4/21/10), 1e-6)

y = np.mean(cnorm[2:nq+2,1:6],1)
dy = np.std(cnorm[2:nq+2,1:6],1)
ax[1].errorbar(qvn2, y, dy, fmt='o', color='k', alpha=.6, label=r'$\beta_0(q)$')

popt2, cov2 = np.polyfit(qvn2, np.log(y), 1, w=y/dy, cov=1)
cov2 = np.sqrt(np.diag(cov2))[0]
ax[1].plot(x2, exp(np.polyval(popt2,x2)), '-', color='k')
#ax[1].legend(loc=3)

x_labels = ax[1].get_xticks()
ax[1].xaxis.set_major_formatter(ticker.FormatStrFormatter('%0.0g'))
c = popt2[0]-popt1[0]
r_loc = np.sqrt(6*(c)/2)
dr_loc = 3/2/r_loc*np.sqrt(cov1**2+cov2**2)
print('\n\nlocalization length: {:.2f} +/- {:.2f} nm'.format(r_loc,dr_loc))


# set style
ax[0].set_xlabel(r'$\mathrm{q}^2$ [$\mathrm{nm}^{-2}]$')
ax[0].set_ylabel(r'$\Gamma [s^{-1}]$')
ax[2].set_xlabel(r'$\mathrm{q}$ [$\mathrm{nm}^{-1}]$')
ax[2].set_ylabel(r'KWW exponent')
ax[1].set_yscale('log')
ax[1].set_xlabel(r'$\mathrm{q}^2$ [$\mathrm{nm}^{-2}]$')
ax[1].set_ylabel(r'contrast')
for axi in ax:
    axi.legend(loc='best')
    # niceplot(axi)
plt.tight_layout()
plt.savefig('um2018/cfpars1.eps')



import os
import re
import glob
import pickle


def make_filename(obj, attr=None, filename=None):

    def splitpath(path):
        if path is None:
            print('No valid file specified.')
        tmp = path.split('/')
        fname = tmp[-1]
        path = os.path.abspath('/'.join(tmp[:-1])) + '/'
        return path, fname

    if attr is None and filename is None:
        raise ValueError(
            'Both attribute name and filename are None. Cannot create filename.')

    savdir = obj.savdir
    if attr is not None:
        orgname = obj.__dict__[attr]
    if filename is None:
        savdir, filename = splitpath(orgname)
    else:
        if '/' in filename:
            filename = os.path.abspath(filename)
            savdir, filename = splitpath(filename)
    return savdir, filename


def mksavdir(sample_name=None, savhome='./', handle_existing='use'):
    """
    Create a directory for saving results.\n
    Usage is: mksavdir(sample_name, savhome, handling_existing)\n
    handling_existig\n
    'use'\tuse same direcotry\n
    'next'\tadd counter
    """
    savhome = os.path.abspath(savhome) + '/'
    if sample_name is None:
        sample_name = input(('Chose a sample name for saving in {}\n'
                             + 'or enter full path.').format(savhome))
    # if '/' in sample_name:
    #     savdir = os.path.abspath(sample_name.rstrip('/')) + '/'
    # else:
    savdir = savhome.rstrip('/') + '/' + sample_name.rstrip('/') + '/'

    try:
        os.makedirs(savdir)
    except FileExistsError:
        if handle_existing == 'use':
            pass
        elif handle_existing == 'next':
            searchstr = '_\d{2}$'
            savsplit = savdir.split('/')
            reg = re.search(searchstr, savsplit[-2])
            folderlist = glob.glob(savdir[:-1]+'_*')
            if reg is None and len(folderlist) == 0:
                savsplit[-2] += '_02'
            else:
                counter = max(list(map(lambda x: int(re.search(searchstr, x).group().lstrip('_')),
                                       folderlist)))
                savsplit[-2] += '_{:02d}'.format(counter+1)
            savdir = '/'.join(savsplit)
            os.makedirs(savdir)
        else:
            print('Could not create directory.')
    savdir = os.path.abspath(savdir) + '/'
    print('Changing savdir to:\n\t{}'.format(savdir))
    return savdir


def save_result(savobj, restype, savdir, filename="", handle_existing='raise', prompt=True):
    savdir = os.path.abspath(savdir)
    filename = re.sub(r'^((' + restype + ')_*)', '', filename)
    filename = re.sub(r'(\.(pkl)*)$', '', filename)
    savname = savdir + '/' + restype + '_' + filename + '.pkl'

    save = True
    if handle_existing == 'next':
        savname = savname.rstrip('.pkl')
        searchstr = '\d{4}$'

        def reg(x): return re.search(searchstr, x.rstrip('.pkl'))
        filelist = glob.glob(savname+'*')
        counter = [int(reg(x).group()) for x in filelist if reg(x) is not None]
        counter.append(-1)
        counter = max(counter) + 1
        savname += '_{:04d}'.format(counter) + '.pkl'
    elif handle_existing in ['overwrite', 'w']:
        pass
    elif handle_existing == 'raise':
        if os.path.isfile(savname) and not prompt:
            raise OSError(('File {} already exists. Change overwrite to ' +
                           'True or choose different name.').format(savname))
        elif os.path.isfile(savname) and prompt:
            user_input = input('File exists. Save anyway? (No/Yes)\t')
            if user_input == 'yes' or user_input == 'Yes':
                pass
            else:
                save = False
    else:
        raise ValueError('{} is not a valid option'.format(handle_existing))

    if save:
        pickle.dump(savobj, open(savname, 'wb'))
        print('\nResults saved to:\n\t{}'.format(savname))
        return savname
    else:
        print('Result has not been saved.')



import numpy as np
from matplotlib import pyplot as plt
from ..Xfit.MCMC_straight_line import mcmc_sl
from ..Xfit.fit_basic import fit_basic
from ..Xfit.FitPoissonGamma import fit_pg
from ..Xplot.niceplot import niceplot
from matplotlib.offsetbox import AnchoredText
from matplotlib import ticker
from ..misc.running_mean import running_mean
from ..misc.resample import resample
from matplotlib.colors import LogNorm
from matplotlib import ticker


def rebin_array(arr, maxlen=500):
    """rebin array and reduce its length to maxlen for generating plots
    """
    l = arr.shape
    if l[0] > maxlen:
        arr = arr[:-(l[0]%maxlen)]
        arr = arr.reshape(maxlen, -1, *l[1:])
        arr = arr.mean(1)
    return arr

def plot_contrast_vs_images(x, y, ax, label=None, marker='o', color='b', ind_avr=False, alpha=0.5):
    """
    """
    ax.plot(x, y, ms=3, label=label, linestyle='', marker=marker, color=color, alpha=alpha)
    ax.set_xlabel('image number')
    ax.set_ylabel(r'$\beta$')
    if ind_avr:
        arrow(x.max(), y.mean(), ax, color)
    return True

def plot_contrast_histogram(x, ax, label=None, color='b', ind_avr=False):
    """
    """
    ax.hist(x, bins=100, density=True, histtype='step', label=label, color=color,)
    ax.set_xlabel(r'$\beta$')    
    ax.set_ylabel(r'$P(\beta)$')
    if ind_avr:
        xx = np.ones(2)*x.mean()
        h = np.histogram(x, bins=100, density=True)
        de = np.diff(h[1][:2])/2
        e = h[1][:-1] + de
        ymax = h[0][np.argmin(np.abs(xx[0]-h[1]+0.5*np.mean(np.diff(h[1]))))]
        yy = np.array([0,ymax])
        ax.plot(xx,yy,'--', color=color)
    return True

def plot_contrast_running_average(x, y, dy, ax, label=None, marker='o', color='b', ind_avr=False, alpha=0.5):
    """
    """
    ax.errorbar(x, y, dy, ms=3, label=label, fmt=marker, color=color, alpha=alpha)
    ax.set_xlabel('image number')
    ax.set_ylabel(r'$\beta$ (run. avr.)')
    return True

def plot_kbar_vs_images(x, y, ax, label=None, marker='o', color='b', npix=False, ind_avr=False, alpha=0.5):
    """
    """
    if npix:
        y = y * npix
        ylab = r'$k_{tot}$'
    else:
        ylab = r'$\langle k\rangle$'
    ax.plot(x, y, label=label, linestyle='', markersize=3, marker=marker, color=color, alpha=alpha)
    ax.set_xlabel('image number')
    ax.set_ylabel(ylab)
    ax.set_yscale('log')
    if ind_avr:
        arrow(x.max(), y.mean(), ax, color)
    return True

def plot_poisson_gamma(x, y, dy, ax, label=None, marker='o', color='b', ind_avr=False):
    """
    """
    ax.errorbar(x, y, yerr=dy, label=label, ms=3, ls='', marker=marker, color=color)
    ax.set_xlabel('k')
    ax.set_ylabel(r'$P(k)$')
    ax.set_yscale('log')
    return True

def plot_contrast_kbar_correlation(x, y, ax, label=None, cmap='inferno', npix=False, ind_avr=False, alpha=0.5):
    """
    """
    if npix:
        x = x * npix
        xlab = r'$k_{tot}$'
    else:
        xlab = r'$\langle k\rangle$'
    ax.hist2d(x, y, bins=50, density=True, norm=LogNorm(), cmap=cmap, label=label, alpha=alpha)
    ax.set_xlabel(xlab)
    ax.set_ylabel(r'$\beta$')
    return True

def plot_contrast_vs_kbar(x, y, ax, label=None, marker='o', color='b', npix=False, ind_avr=False, alpha=0.5):
    """
    """
    if npix:
        x = x * npix
        xlab = r'$k_{tot}$'
    else:
        xlab = r'$\langle k\rangle$'
    ax.plot(x, y, linestyle='', marker=marker, markersize=2, color=color, label=label, alpha=alpha)
    ax.set_xlabel(xlab)
    ax.set_ylabel(r'$\beta$')
    return True

def plot_prob_vs_kbar(x, y, ax, probk, label=None, marker='o', color='b', npix=False, ind_avr=False, alpha=0.5):
    """
    """
    if npix:
        x = x * npix
        y = y * npix
        xlab = r'$k_{tot}$'
        ylab = r'n(%d)' % probk
    else:
        ylab = r'P(%d)' % probk
        xlab = r'$\langle k\rangle$'
    ax.plot(x, y, linestyle='', marker=marker, markersize=2, color=color, label=label, alpha=alpha)
    ax.set_xlabel(xlab)
    ax.set_ylabel(ylab)
    return True

def arrow(x, y, ax, color):
    c1 = (x,y)
    c2 = (x*1.1,y)
    arprp = dict(arrowstyle="-|>", connectionstyle='arc3', color=color,)
    ax.annotate("", xy=c1, xycoords='data', xytext=c2, textcoords='data', arrowprops=arprp)
    
def plot_quicklook(obj, plot, idx, nq, ax=None, c0=0, ratio=0, maxlen=500, format_ticks=False,
                   ind_avr=True, lfs=10, total_counts=False, return_respfnc=False, probk=2,
                   fit_pg=True, alpha=0.5, **kwargs):
    """Plot quick overview of speckle contrast, mean photon counts and Probability Distribution
    """
    contrast = obj.contrast[0][idx]
    prob = obj.prob[0][idx][1:,1:]

    for i,j in enumerate(nq):
        ci = i + c0*len(nq) # index for colors and markers
        color = obj.colors[ci]
        marker = obj.markers[idx%len(obj.markers)]
        if total_counts:
            npix = obj.Xana.setup['gproi'][j]
        else:
            npix = False
        
        if i == 0:
            labstr = 't_exp: {:.2e}s'.format(obj.t_exposure[idx])
        else:
            labstr = None
            
        b = contrast[j,ratio+1]
        xv_b = np.where(~b.mask)[0]
        x_b = rebin_array(xv_b, maxlen)
        br = rebin_array(b[xv_b], maxlen)

        bt = running_mean(b[xv_b])
        bt = rebin_array(bt, maxlen)

        kb = contrast[j,0]
        xv_kb = np.where(~kb.mask)[0]
        kb = rebin_array(kb[xv_b], maxlen)
        p = prob[j,probk+1][xv_b]

        pp = np.mean(prob[j,1:],-1)
        ind = np.where(pp)[0]
        dpp = np.std(prob[j,1:],-1)
        k = np.arange(pp.size)

        if plot == 'bvi':
            labmean = '%.2f' % br.mean()
            plot_contrast_vs_images(x_b, br, ax, labmean, marker, color, ind_avr, alpha)
        elif plot == 'pbb':
            plot_contrast_histogram(b[~b.mask], ax, labstr, color, ind_avr)
        elif plot == 'brvi':
            plot_contrast_running_average(x_b, bt[:,0], bt[:,1], ax, labstr, marker, color, alpha)
        elif plot == 'kbvi':
            labmean = '%.3g' % kb.mean()
            plot_kbar_vs_images(x_b, kb, ax, labmean, marker, color, npix, ind_avr, alpha)
        elif plot == 'bvkb':
            plot_contrast_vs_kbar(kb, br, ax, labstr, marker, color, npix, alpha)
        elif plot == 'pbk':
            plot_poisson_gamma(k[ind], pp[ind], dpp[ind], ax, labstr, marker, color)
            # if fit_pg:
            #     pb_pg = np.hstack((np.zeros(k[ind].size)))
            #     fitpg()
        elif plot == 'pkvkb':
            plot_prob_vs_kbar(kb, p, ax, probk, labstr, marker, color, npix, alpha)
   
    if ind_avr:
        ax.legend(loc='best')
    # niceplot(ax, autoscale=True, grid=False, lfs=lfs)

# def plot_poisson_gamma(obj, idx, nq, dofit=True):
#     """Show the photon statistics by plotting the Poisson-Gamma distribution
#     """
#         if data == 'original' or self.corrFuncRescaled is None:
#             corrFunc = list(self.corrFunc)
#         elif data == 'rescaled':
#             corrFunc = list(self.corrFuncRescaled)
#         else:
#             raise ValueError('No usable correlation data defined.')

#         if nq is None:
#             pass
#         elif type(nq) == int:
#             self.nq = np.arange(nq)
#         else:
#             self.nq = nq

#         if color_mode == 0:
#             color_multiplier, color_repeater = self.nq.size, len(corrFunc)
#         elif color_mode == 1:
#             color_multiplier, color_repeater = self.nq.size*len(corrFunc), 1

#         self.update_colors(cmap, color_multiplier, color_repeater)
#         self.update_markers( len(corrFunc), change_marker)

#         self.g2plotl = [[]] * len(corrFunc)
#         self.pars = [[]] * len(corrFunc)
#         self.fit_result = [[[] for i in range(self.nq.size)]] * len(corrFunc) # possible bug: will cause
#                                                                               # wrong references
#         ci = 0
#         for j, (cfi, dcfi) in enumerate(corrFunc):
#             rates = np.zeros((self.nq.size, 3*nmodes+3, 2))
#             ti = cfi[1:,0]
#             for i,qi in enumerate(self.nq):
#                 if i == 0:
#                     cf_id = self.db_id[j]
#                 else:
#                     cf_id = None
#                 res = fitg2(ti, cfi[1:,qi+1], err=dcfi[1:,qi+1], qv=self.Xana.setup['qv'][qi],
#                             ax=ax, color=self.colors[ci], dofit=True,
#                             marker=self.markers[j%len(self.markers)], cf_id=cf_id,
#                             modes=nmodes, **kwargs)
#                 self.fit_result[j][i] = res[2:4]
#                 self.g2plotl[j].append(list(itertools.chain.from_iterable(res[4])))
#                 if dofit: 
#                     if i == 0:
#                         db_tmp = self.init_pars(list(res[2].params.keys()))
#                     entry = [cfi[0,qi+1], *res[0].flatten(), *res[1]]
#                     db_tmp.loc[i] = entry
#                 else:
#                     db_tmp = 0
#                 ci += 1
#             self.pars[j] = db_tmp
#         plt.tight_layout()

        
#     for j,ei in zip(tind,extf[tind]):
#         nn = np.ceil(len(roi_list)/2)
#         f, ax = plt.subplots(int(nn), 2, figsize=(9,nn*4))
#         f.suptitle('exposure time = {:.2g}s'.format(ei))
#         ax = ax.ravel()
  
#         for l,roii in enumerate(roi_list):
#             kv, kb, dkb, pba, dpba = selectdata(xsvs, ei, t_int=ext, dxsvs=xsvs_err, roi_idx=roii, **psel)

#             out = fpg.fitpg_iterative(kb, pba, kv, pberr=1/pba, **pfit)
#             M[roii,0,:] = qv[roii]
#             M[roii,j+1,:3] = (out[0],out[1],np.mean(out[2]))

#             plotpoissongamma(out[2], pba, kv, out[0], dkb, dpba*0, out[1],
#                              confint=(1/np.mean(cnorm[roii+2,np.where(cnorm[0,1:]==ei)]),),
#                              q=qv[roii], ax=ax[l], cmap='Set1',**pfit)

#         if doplot:
#             plt.tight_layout()




import numpy as np

def map2ct(mat):
    ct = np.zeros_like(mat)*np.nan
    for i in range(1,mat.shape[1]):
        ct[:-i,i] = mat[i:,i]
    return ct

def map2g2(mat):
    mat = np.rot90(mat,1)
    n2 = mat.shape[1]
    for i in range(-n2,n2):
        if i == -n2:
            ct = np.ones((int(n2/2),2*n2))*np.nan
        tmp = np.diag(mat,i)
        tmp = tmp[int(tmp.size/2):]
        ct[:tmp.size,i+n2] = tmp
    return ct


import numpy as np

def resample( x, y, dy=None, npoints=100, log=True, method=0):
    """Resample data.
    """
    x = np.ma.masked_where(np.isnan(x), x)
    y = np.ma.masked_where(np.isnan(y), y)

    if y.ndim == 1:
        y = np.expand_dims(y, 1)
        if dy is not None:
            dy = np.expand_dims(dy, 1)

    if dy is not None:
        mask = (dy<=0) | np.isnan(y)
        y = np.ma.masked_array(y, mask=mask)
        dy = np.ma.masked_array(dy, mask=mask)
        returned = True
    else:
        returned = False
        dy = np.ones_like(y)

    if log:
        newx = np.logspace(np.log10(x.min()), np.log10(x.max()), npoints+1)
    else:
        newx = np.linspace(x.min(), x.max(), npoints+1)

    ind = np.digitize(x.filled(0), newx)
    un_ind = np.unique(ind)
    newx = np.zeros(un_ind.size)
    newy = np.zeros((newx.size, y.shape[-1]))
    newdy = newy.copy()
    for j, i in enumerate(un_ind):
        ii = np.where((ind==i))[0]
        if y[ii].filled(0).sum() == 0:
            newx[j] = np.nan
            newy[j] = np.nan
            continue
        if method:
            m = np.nanmean(y[ii],0)
            tmp = (m[None,:]-y[ii])**2 / m[None,:]**2
            dy[ii][tmp>0] = tmp[tmp>0]
        newx[j] = np.ma.average(x[ii],)
        newy[j], newdy[j] = np.ma.average(y[ii], weights=1/dy[ii]**2, axis=0, returned=1)
    # newx = newx[1:]
    # newy = newy[1:]
    # newdy = newdy[1:]
    newdy[newdy>0] = np.sqrt(1/newdy[newdy>0])
    if returned:
        return newx, newy, newdy
    else:
        return newx, newy





import numpy as np
import numpy.ma as ma
from Xdrop.dropletizem import dropimgood_sel
import time
import pickle
import re
from matplotlib import pyplot as plt


def dropletizedata(data, pars, mask=None, dark=None, savdir='./', savname=None):
    #print(dropimgood_sel.__doc__)
    dim = data.shape
    mshape = dim[-2:]
    data = data.reshape(-1,*mshape)
    datdrop = np.zeros(data.shape, dtype=np.uint16)

    if dark is None:
        dark = np.zeros(mshape)

    if mask is None:
        mask = np.ones(mshape, dtype=np.bool)

    pix  = []
    for imgn in range(data.shape[0]):
        img = data[imgn].astype(int)
        imd = dropimgood_sel(img, dark, ~mask,
                             pars['background'],
                             pars['lower_threshold'],
                             pars['upper_threshold'],
                             pars['number_photons'],
                             pars['adusPphoton'],
                             pars['pixelPdroplet'])

        datdrop[imgn] = imd[-1]
#        pix.append(imd[:-1])
    if savname is not None:
        np.save(savdir + savname + '_dropletized.npy', datdrop)
        f = open(savdir + savname + '_pix.pkl', 'wb')
        pickle.dump({'pix':pix},f)
        f.close()

    datdrop = np.squeeze(datdrop.reshape(dim))
    return datdrop

def testDropletizing(data, pars, dark=None, mask=None):
    im = data.copy()
    if 'roi' in pars.keys():
        roi = pars['roi']
        xl = roi[0]
        yl = roi[1]
    else:
        xl = [0, dark.shape[1]]
        yl = [0, dark.shape[0]]
    im = im[yl[0]:yl[1],xl[0]:xl[1]]
    if dark is not None:
        dark = dark[yl[0]:yl[1],xl[0]:xl[1]]
        im = im - 1.*dark
    if mask is not None:
        mask = mask[yl[0]:yl[1],xl[0]:xl[1]]
        im *= mask
    nx = im.shape[1]
    ny = im.shape[0]
    pars['nx'] = nx
    pars['ny'] = ny
    imd1 = dropletizedata(im, pars, mask=mask, dark=dark)
        
    tstr = ['raw image', 'parameters 1', 'parameters 2']
    fig, ax = plt.subplots(1,2,figsize=(9,6))
    ax = ax.ravel()
    im[im<pars['background']] = 0
    for i,imp in enumerate([im, imd1,]):
        pl = ax[i].imshow(imp, interpolation='nearest', cmap='Blues')
        ind_x = np.arange(xl[1]-xl[0])
        ind_y = np.arange(yl[1]-yl[0])
        x, y = np.meshgrid(ind_x, ind_y)
        for xi, yi in zip(x.flatten(), y.flatten()):
            if imp[yi,xi]>0:
                c = '{0}'.format(int(imp[yi,xi]))
                ax[i].text(xi, yi, c, va='center', ha='center', fontsize=8, color='w', fontweight='bold')

        ax[i].set_xticks(ind_x-.5)
        ax[i].set_yticks(ind_y-.5)
        ax[i].set_xticklabels([])
        ax[i].set_yticklabels([])
        ax[i].set_xlim(ind_x[0]-.5, ind_x[-1]+.5)
        ax[i].set_ylim(ind_y[0]-.5, ind_y[-1]+.5)

    I = im.sum()
    ph = imd1.sum()
    ax[0].set_title(r'$I={:.00f}\,adus\, ({:.00f}ph$)'.format(I,I/pars['adusPphoton']))
    ax[1].set_title(r'$I={:.00f}\,ph$'.format(ph))
    plt.tight_layout(pad=2)
    '''
    number = 0
    if saveimages:
        while os.path.isfile(savdir + savname):
            number = int(re.findall(r'\d+', savname)[0]) + 1
            savname = 'pic_{0:02}.pdf'.format(number)
        plt.savefig(savdir + savname)#, dpi=300)
    '''
    plt.show()




#! /usr/bin/env python
import numpy as np
import lmfit
from .PoissonGammaDistribution import PoissonGamma as poisgam

def fit_pg_likelihood(prob, npix, err=None, kv=None, init={}, fix=None, method='Nelder-mead'):
    """ Fit the Poisson-Gamma distribution using the likelihood ratio approach.
    """
    if kv is None:
        kv = np.arange(prob.shape[0]-2)

    kb = prob[1]
    prob = prob[kv+2]

    #make initial guess for parameters
    for vn in ['M']:
        if vn not in init.keys():
            if vn == 'M':
                init[vn] = (4, 1, None)

    # initialize fit parameters
    pars = lmfit.Parameters()
    pars.add('M', value=init['M'][0], min=init['M'][1], max=init['M'][2])
                             
    if fix is not None:
        for vn in fix.keys():
            pars[vn].set(value=fix[vn], vary=0)

    if err is not None:
        err = np.abs(err)
        wgt = err.copy()
        wgt[wgt>0] = 1./wgt[wgt>0]**2
    else:
        wgt = None
        
    # independent variable is <k>
    def likelihood_ratio(pars, prob, eps=None):
        """2D Residual function to minimize
        """
        v = pars.valuesdict()
        
        chi2 = 0
        for i in range(kv.size):
            probi = prob[i]
            ind = np.where(probi)[0]
            pg = poisgam(kb[ind], v['M'], kv[i], ind_var='kb')
            chi2 += np.sum(probi[ind] * np.log(pg/probi[ind]))
        chi2 *= -2 * npix
        return chi2
    
    out = lmfit.minimize(likelihood_ratio, pars, args=(prob,), kws={'eps':wgt}, method=method,
                         nan_policy='omit')

    pars_arr =  np.zeros((1,2))
    for i,vn in enumerate(['M']):
        pars_arr[i,0] = 1./out.params[vn].value
        # pars_arr[i,1] = pars_arr[i,0]**2*out.params[vn].stderr
    gof = np.array([out.chisqr, out.redchi, out.bic, out.aic])

    return pars_arr, gof, out, lmfit.fit_report(out)




import numpy as np
from time import time


def mp_prob(method, nbins, nf, lind, nq, quc, quce):

    def xsvs_full(chunk):
        for qi in xnq:
            roi = chunk[qi]
            for i,line in enumerate(roi):
                tmp = np.append(np.sum(line), np.histogram(line, bins=np.arange(nbins+1)-0.5)[0])
                prob[qi,:,t0+i] = tmp / lind[qi]

    tcalc = time()
    xnq = range(nq)
    prob = np.zeros((nq,nbins+1,nf), dtype=np.float32)

    t0 = 0
    while t0 < nf:
        chunk = quc.get()
        chunk_size = chunk[0].shape[0]
        xsvs_full(chunk)
        t0 += chunk_size

    # END OF MAIN LOOP put results to output queue
    quc.close()
    quc.join_thread()
    tcalc = time() - tcalc
    quce.put([prob, tcalc])



import numpy as np
from scipy.special import gamma, gammaincc


def V2(x, **p):
    t = p['t']
    g = p['g']
    return 2**((g-2)/g)*t/(g*x**2)*(2**(1/g)*x*(gamma(1/g)-gamma(1/g)*gammaincc(1/g,2*(x/t)**g))
                                    + t*(gamma(2/g)*gammaincc(2/g,2*(x/t)**g)-gamma(2/g)))

def V2a(x, **p):
    a = p['a']
    b = p['b']
    return a + b*V2(x, **p)

'''
def M(x, **p):
    return 1/bandy_V2(x, **p)

def bandy_Ma(x, **p):
    return 1/bandy_V2a(x, **p)

def bandy_logMa(x, **p):
    return log10(1/bandy_V2a(x, **p))

def bandy_logV2(x, *p):
    return log10(bandy_V2(x, *p))

def bandy_logM(x, **p):
    return log10(bandy_M(x, **p))

def bandy_Corr(x, **p):
    c = p['c']
    return bandy_M(x, **p) - 1/c

def bandy_logCorr(x, **p):
    c = p['c']
    return bandy_M(x, **p) - 1/c
'''



import numpy as np
import multiprocessing as mp
from queue import PriorityQueue
from XsvsAna.pyxsvs3 import pyxsvs
from misc.xsave import save_result
from multiprocessing.managers import SyncManager
from helper import *

class MyManager(SyncManager):
    pass

def Manager():
    m = MyManager()
    m.start()
    return m

def xsvs(obj, series_id, first=0, last=None, filename="", method='full', nbins=15,
         handle_existing='next', nprocs=4, nread_procs=4, chunk_size=100, dark=None,
         load_verbose=False, dtype=np.float32, **kwargs):

    #MyManager.register("PriorityQueue", PriorityQueue)  # Register a shared PriorityQueue

    for sid in series_id:
        print('\n\n#### Starting XSVS Analysis ####\nSeries: {} in folder {}\n'.format(sid, obj.xdata.datdir))

        if last is None:
            last = obj.xdata.meta.loc[sid,'nframes']

        if dark is not None:
            if type(dark)==int:
                print('Loading DB entry {} as dark.'.format(dark))
                dark = obj.get_item(dark)['Isaxs']

        read_data_opt = {'first':(first,),
                         'last':(last,),
                         'dark':dark,
                         'verbose':load_verbose,
                         'dtype':dtype,
                         'qsec':obj.setup['qsec'],
                         'output':'2dsection',
                         'nprocs':1
        }
        read_data_opt.update(kwargs)

        qsec = obj.setup['qsec']
        xpcs_opt = {'nimages':last-first,
                    'dim':(qsec[1][0]-qsec[0][0]+1,qsec[1][1]-qsec[0][1]+1)
        }

        t_e = 0
        for attr in ['t_exposure', 'pulseLength']:
            if attr in obj.xdata.meta.columns:
                item = obj.xdata.meta.loc[sid,attr]
                if attr == 'pulseLength':
                    t_e += item * 1e15
                else:
                    t_e += item

        fmax = obj.xdata.meta.loc[sid,'nframes']
        chunks = [np.arange(first+i*chunk_size,min([min(first+(i+1)*chunk_size,last),fmax]))
                  for i in range(np.ceil((last-first)/chunk_size).astype(np.int32))]
 
        print('Using {} processes to read data.'.format(nread_procs))
#        m = Manager()
        # dataQ = m.PriorityQueue(nread_procs)
        # indxQ = m.PriorityQueue()
        dataQ = mp.Queue(nread_procs)
        indxQ = mp.Queue()
        read_data_opt['method'] = 'queue_chunk'
        read_data_opt['dataQ'] = dataQ
        read_data_opt['indxQ'] = indxQ
        xpcs_opt['dataQ'] = dataQ

        for i,chunk in enumerate(chunks):
            indxQ.put((i,chunk))

        lock = 0
        if 'h5' in obj.xdata.fmtstr:
            lock = mp.Lock()
            read_data_opt['lock'] = lock

        procs = []
        for ip in range(nread_procs):
            procs.append(mp.Process(target=obj.xdata.get_series,
                                    args=(sid,), kwargs=read_data_opt))
            procs[ip].start()

        probd = pyxsvs(xpcs_opt, obj.setup['qroi'], method=method, nbins=nbins, t_e=t_e,
                       qv=obj.setup['qv'], nprocs=nprocs, qsec=obj.setup['qsec'][0])

        # stopping processes
        for ip in range(nread_procs):
            procs[ip].join()

        # closing queues 
        dataQ.close()
        dataQ.join_thread()
        indxQ.close()
        indxQ.join_thread()

        f = obj.xdata.datdir.split('/')[-2] + '_s' + str(obj.xdata.meta.loc[sid, 'series']) + filename
        savfile = save_result(probd, 'xsvs', obj.savdir, f, handle_existing=handle_existing)
        obj.add_db_entry(sid, savfile)



import numpy as np

def mat2evt(roi):
    '''Function to convert matrix of npix x ntimes
       into a vector of events
    '''
    ntimes, npix = roi.shape
    s = np.sum(roi,1).astype(np.int32)
    t = np.zeros(s.sum(),np.int32)
    s0 = 0
    css = np.append(0,np.cumsum(s))
    for i in range(css.size-1):
        t[css[i]:css[i+1]] = i
    pix = t.copy()

    eventpix = np.tile(np.arange(npix),ntimes).astype(np.int32)
    roi = roi.ravel() 
    ind = np.where(roi)
    roi = roi[ind]
    csroi = np.append(0,np.cumsum(roi)).astype(np.int32)
    eventpix = eventpix[ind]
    for i in range(len(csroi)-1):
        pix[csroi[i]:csroi[i+1]] = eventpix[i]

    return pix, t, s

def cftomt(d, par=16, err2=None):
    '''Function to bin the correlation function. Returns correlation
       functions similar to the multi tau approach.
    '''
    nt = []
    nd = []
    er = []

    if err2 is None:
        err2 = np.ones(d.shape[0], dtype=d.dtype)

    err2 = np.ma.masked_array(err2)

    for i in range(par):
        nt.append(d[i,0])
        nd.append(d[i,1])
        er.append(np.sqrt(err2[i]))

    t = d[par:,0]
    val = d[par:,1]
    err2 = err2[par:]

    while len(t) >= par:
        invsvar = 1/(1/err2[:-1] + 1/err2[1:])
        nval = (val[:-1]/err2[:-1] + val[1:]/err2[1:]) * invsvar
        tt = (t[:-1] + t[1:])/2
        for i in range(0,par,2):
            nt.append(tt[i])
            nd.append(nval[i])
            er.append(np.sqrt(invsvar[i]))
        t = tt[par:-1]
        val = nval[par:-1]
        err2 = invsvar[par:-1]

    # for i in range(len(t)):
    #     nt.append(t[i])
    #     nd.append(val[i])
    #     er.append(err2[i])

    x = np.array([nt,nd,er]).T  
    return x




from .pyxpcs3 import pyxpcs
from .eventcorrelator3 import eventcorrelator

class Xpcs:

    def __init__():

        pass

    def pyxpcs(*args, **kwargs):
        return pyxpcs(*args, **kwargs)

    def eventcorrelator(*args, **kwargs):
        return eventcorrelator(*args, **kwargs)



import numpy as np
from matplotlib import pyplot as plt
import matplotlib
import itertools
import copy
import pandas as pd

from ..Decorators import Decorators
from ..Analist import AnaList
from .PlotContrast import plot_quicklook
from ..Xplot.niceplot import niceplot
from .PhotonStats import prob2beta, average_beta, beta_from_likelihood, prob2betasigma


class VisbFunc(AnaList):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ratesl = None
        self.fit_result = None
        self.pars = None
        self.v2 = None
        self.prob = None
        self.t_exposure = None
        self.contrast = None
        self.v2plotl = None
        self.nq = np.arange(len(self.Xana.setup.qroi))
        self.db_id = None

    def __str__(self):
        return 'Corr Func class for g2 displaying.'

    def __repr__(self):
        return self.__str__()
        
    @Decorators.input2list
    def get_prob(self, db_id, merge=False, verbose=True):
        """Load photon probabilities from database
        """
        self.db_id = db_id
        self.prob =  [[] for _ in range(2)]
        t_exposure = np.empty(len(db_id))
        l = []
        for i, sid in enumerate(db_id):
            try:
                d = self.Xana.get_item(sid)
                l.append(d['prob'])
                t_exposure[i] = d['prob'][0,0,0]
            except KeyError:
                print('Could not load item {}'.format(sid))

        ind = np.argsort(t_exposure)
        self.t_exposure = t_exposure[ind]
        self.prob[0] = [l[i] for i in ind]

        if verbose:
            print('Loaded probabilities of {} series.'.format(len(db_id)))
       
        self.prob[1] = copy.deepcopy(self.prob[0])

        if merge:
            self.merge_problist()
        
    def merge_problist(self):
        """merge lists of photon probabilities
        """
        self.t_exposure, cnt = np.unique(self.t_exposure, return_counts=True)
        cnt = np.cumsum(cnt)
        cnt = np.append(0,cnt)
        
        for ii, probl in enumerate(self.prob):
            prob = []
            for i,j in zip(cnt[:-1],cnt[1:]):
                prob.append(np.concatenate(probl[i:j],-1))
            self.prob[ii] = prob

    def calculate_contrast(self, method='formula', sigma=3):
        """Calculate speckle contrast in different ways.
        """
        self.contrast = [[] for _ in range(2)]
        for ii, probl in enumerate(self.prob[:1]):
            for prob in probl:
                if method == 'formula':
                    gproi = self.Xana.setup.gproi
                    self.contrast[ii].append(prob2beta(prob, gproi))
                    # self.contrast[ii+1].append(prob2betasigma(prob, gproi, sigma))

    def calculate_v2(self, ratio=0, **kwargs):
        """Calculate Bandy function in different ways.
        """
        methods = ['av', 'md', 'mx', 'pg', 'ml']
        self.v2 = {}
        v2_av = average_beta(self.t_exposure, self.Xana.setup.qv, self.contrast, ratio)
        for i,m in enumerate(methods[:3]):
            self.v2[m] = v2_av[i]

        self.v2['ml'] = beta_from_likelihood(self.t_exposure, self.Xana.setup.qv, self.prob,
                                             self.Xana.setup.gproi, **kwargs)
 
        return None

    def quicklook(self, plot='default', idx=None, nq=None, ax=None, color_mode=0, change_marker=0,
                  cmap='magma', *args, **kwargs):
        """Plot overview of photon probabilities and speckle contrast
        """
        if plot == 'default':
            plot = ['bvi', 'brvi', 'kbvi', 'bvkb', 'pbb', 'pbk']
        npl = len(plot)

        if nq is None:
            nq = self.nq

        if idx is None:
            idx = range(len(self.contrast[0]))
        nlines = len(idx)

        if ax is None:
            n = int(np.ceil(npl/2))
            fig, ax = plt.subplots(n, int(npl>1)+1, figsize=(5+4*(npl>1),3.5*n))

        if type(ax) == np.ndarray:
            ax = ax.flatten()
        else:
            ax = [ax,]

        ax_idx = np.arange(npl,dtype=np.int16)
        c_idx = np.arange(nlines)
            
        if color_mode == 0:
            color_multiplier, color_repeater = nq.size, nlines
        elif color_mode == 1:
            color_multiplier, color_repeater = nq.size * nlines, 1
        self.update_colors(cmap, color_multiplier, color_repeater)
        self.update_markers(nlines, change_marker)

        for i,p in enumerate(plot):
            for j,jj in enumerate(idx):
                plot_quicklook(self, p, jj, nq, ax[ax_idx[i]], c0=c_idx[j], *args, **kwargs)
        plt.tight_layout()
    
    @Decorators.init_figure()
    def plot_g2(self, nq=None, err=True, ax=None, nmodes=1, data='original', cmap='magma',
                    change_marker=False, color_mode=0, dofit=True, **kwargs):
        
        if data == 'original' or self.corrFuncRescaled is None:
            corrFunc = list(self.corrFunc)
        elif data == 'rescaled':
            corrFunc = list(self.corrFuncRescaled)
        else:
            raise ValueError('No usable correlation data defined.')

        if nq is None:
            pass
        elif type(nq) == int:
            self.nq = np.arange(nq)
        else:
            self.nq = nq

        if color_mode == 0:
            color_multiplier, color_repeater = self.nq.size, len(corrFunc)
        elif color_mode == 1:
            color_multiplier, color_repeater = self.nq.size*len(corrFunc), 1

        self.update_colors(cmap, color_multiplier, color_repeater)
        self.update_markers( len(corrFunc), change_marker)

        self.g2plotl = [[]] * len(corrFunc)
        self.pars = [[]] * len(corrFunc)
        self.fit_result = [[[] for i in range(self.nq.size)]] * len(corrFunc) # possible bug: will cause
                                                                              # wrong references
        ci = 0
        for j, (cfi, dcfi) in enumerate(corrFunc):
            rates = np.zeros((self.nq.size, 3*nmodes+3, 2))
            ti = cfi[1:,0]
            for i,qi in enumerate(self.nq):
                if i == 0:
                    cf_id = self.db_id[j]
                else:
                    cf_id = None
                res = fitg2(ti, cfi[1:,qi+1], err=dcfi[1:,qi+1], qv=self.Xana.setup.qv[qi],
                            ax=ax, color=self.colors[ci], dofit=True,
                            marker=self.markers[j%len(self.markers)], cf_id=cf_id,
                            modes=nmodes, **kwargs)
                self.fit_result[j][i] = res[2:4]
                self.g2plotl[j].append(list(itertools.chain.from_iterable(res[4])))
                if dofit: 
                    if i == 0:
                        db_tmp = self.init_pars(list(res[2].params.keys()))
                    entry = [cfi[0,qi+1], *res[0].flatten(), *res[1]]
                    db_tmp.loc[i] = entry
                else:
                    db_tmp = 0
                ci += 1
            self.pars[j] = db_tmp
        plt.tight_layout()
        plt.show()

    @staticmethod
    def init_pars(names):
        # names = [names[i//2] if (i+1)%2 else '+/-' for i in range(len(names)*2)]
        names = [names[i//2] if (i+1)%2 else 'd'+names[i//2] for i in range(len(names)*2)]
        names.insert(0, 'q')
        names.extend(['chisqr', 'redchi', 'bic', 'aic'])
        return pd.DataFrame(columns=names)

    def reset_rescaled(self):
        """Reset rescaled probabilities by copying the original ones.
        """
        self.prob[1] = copy.deepcopy(self.prob[0])
        
    @Decorators.init_figure()
    @Decorators.input2list
    def plot_trace(self, db_id, log='', ax=None):
        axtop = ax.twiny()

        ci = 0
        for sid in db_id:
            corfd = self.Xana.get_item(sid)
            trace = corfd['trace']
            framen = np.arange(trace.shape[0])
            time = corfd['twotime_xy']
            for i, iq in enumerate(self.nq):
                ax.plot(time, trace[:,iq], color=self.colors[ci])
                axtop.plot(framen, trace[:,iq], color=self.colors[ci])
                ci += 1

        ax.set_ylabel('photons per pixel')
        ax.set_xlabel('time in [s]')
        # niceplot(ax, autoscale=0)
        axtop.set_xlabel('frame number')
        # niceplot(axtop, autoscale=False, grid=False)
        if 'logx' in log:
            ax.set_xscale('log')
        if 'logy' in log:
            ax.set_yscale('log')
        
        # ax.set_title('trace', fontweight='bold', fontsize=14, y=1.14)

    def get_static_contrast(self, *args, **kwargs):
        self.staticContrast = staticcontrast(self, *args, **kwargs)

    def g2_totxt(self, savname):
        return 0



import numpy as np

def running_mean(x):
    """
    Calculate running average
    Initialize M1 = x1 and S1 = 0.

    For subsequent x‘s, use the recurrence formulas

    Mk = Mk-1+ (xk – Mk-1)/k
    Sk = Sk-1+ (xk – Mk-1)*(xk – Mk).

    For 2 ≤ k ≤ n, the kth estimate of the variance is s2 = Sk/(k – 1).
    """
    out = np.zeros((x.size,3))
    for i in range(x.size):
        if i == 0:
            out[i,:] = (x[i],0,0)
        else:
            Mi = out[i-1,0] + (x[i] - out[i-1,0])/(i+1)
            Si = out[i-1,2] + (x[i] - out[i-1,0])*(x[i] - out[i,0])
            out[i,:] = (Mi,np.sqrt(np.abs(Si)/i**2),Si)
    return out[:,:2]

def runavr(x):
    out = np.zeros((x.size-1,3))
    for i in range(x.size-1):
        if i == 0:
            out[i,:] = (x[i],0,0)
        else:
            Mi = out[i-1,0] + (x[i] - out[i-1,0])/(i+1)
            Si = out[i-1,2] + (x[i] - out[i-1,0])*(x[i] - out[i,0])
            out[i,:] = (Mi,np.sqrt(Si/i**2),Si)
    return out



import numpy as np
import lmfit
from .g2function import g2
import matplotlib
from matplotlib import pyplot as plt
from ..Xplot.niceplot import niceplot

def fitg2( t, cf, err=None, mode='semilogx', modes=1, init={}, fix={}, dofit=False,
           doplot=False, marker='o', qv=None, h_plot=None, ax=None, xl=None, yl=None,
           color=None, alpha=1., markersize=3., cf_id=None, data_label=None, errthr=1e-4):

    # define residual function
    def res(pars, x, data=None, eps=None):
        """2D Residual function to minimize
        """
        v = pars.valuesdict()
        for i in range(modes):
            if i == 0:
                model = g2(x, t=v['t{}'.format(i)], b=v['b{}'.format(i)],
                           g=v['g{}'.format(i)], a=v['a'])
            else:
                model += g2(x, t=v['t{}'.format(i)], b=v['b{}'.format(i)],
                            g=v['g{}'.format(i)], a=0)
        if eps is not None:
            resid = np.abs(data - model)*np.abs(eps)
        else:
            resid = np.abs(data - model)
        return resid
    
    #make initial guess for parameters
    for i in range(modes):
        for s in 'tgb':
            vn = s+'{}'.format(i)
            if vn not in init.keys():
                if s == 't':
                    t0 = np.logspace(np.log10(t.min()), np.log10(t.max()), modes+2)[i+1]
                    init[vn] = (t0, -np.inf, np.inf)
                elif s == 'g':
                    init[vn] = (1, .2, 1.8)
                elif s == 'b':
                    init[vn] = (.1, 0, 1)
    if 'a' not in init.keys():
        init['a'] = (np.mean(cf[-10:]), 0, 2)
    if 'beta' not in init.keys():
        init['beta'] = (0.2, 0, 1)

    #initialize parameters
    pars = lmfit.Parameters()
    for i in range(modes):
        for s in 'tgb':
            vn = s + '{}'.format(i)
            pars.add(vn, value=init[vn][0], min=init[vn][1], max=init[vn][2], vary=1)
    pars.add('a', value=init['a'][0], min=init['a'][1], max=init['a'][2], vary=1)
    pars.add('beta', value=init['beta'][0], min=init['beta'][1], max=init['beta'][2], vary=1)

    if modes > 1:
        beta_constraint = 'a+beta-1-'+'-'.join(['b{}'.format(x) for x in range(1,modes)])
        pars['b0'].set(expr=beta_constraint)
    else:
        pars['b0'].set(expr='beta')

    if err is not None:
        inderr = np.isfinite(err) & (err>0) 
    else:
        inderr = np.ones_like(cf, bool)

    ind = np.where(np.isfinite(cf) & inderr)[0]
    ind = ind[:-2]
    t = t[ind]
    cf = cf[ind]
    if err is not None:
        err = err[ind]
    
    for vn in fix.keys():
        if vn in pars.keys():
            pars[vn].set(value=fix[vn], min=-np.inf, max=np.inf, vary=0)

    if err is not None:
        wgt = err.copy()
        wgt = 1./wgt
    else:
        wgt = None

    if 'semilogx' in mode:
        if err is not None:
            wgt = np.log10(wgt)
        elif err == 't':
            wgt = 1./np.log10(t)
        else:
            wgt = None

    if 'y' in mode:
        if err is not None:
            wgt = cf-1
            
    if dofit:
        out = lmfit.minimize(res, pars, args=(t,), kws={'data':cf, 'eps':wgt},
                             nan_policy='omit')

    # do all the plotting
    pl = []
    if doplot:
        if xl is None:
            if ax is None:
                ax = plt.gca()
            xl = ax.get_xlim()
            if xl[0] == 0:
                xl = (np.min(t)*0.5,np.max(t)*1.5)
            
        xf = np.logspace(np.log10(xl[0]),np.log10(xl[1]),50)
        if dofit:
            for i in range(modes):
                if i == 0:
                    g2f = g2(xf, t=out.params['t{}'.format(i)].value,
                             b=out.params['b{}'.format(i)].value,
                             g=out.params['g{}'.format(i)].value,
                             a=out.params['a'].value)
                else:
                    g2f += g2(xf, t=out.params['t{}'.format(i)].value,
                              b=out.params['b{}'.format(i)].value,
                              g=out.params['g{}'.format(i)].value, a=0)

        if 'legf' in doplot and dofit:
            pard = {'t':r'$t: {:.2e}\mathrm{{s}},\,$',
                    'g':r'$\gamma: {:.2g},\,$',
                    'b':r'$\mathrm{{b}}: {:.3g},\,$',
                    'a':r'$\mathrm{{a}}: {:.3g},\,$'}
            labstr_fit = ''
            for i in range(modes):
                for vn in 'tgba':
                    if vn == 'a' and i > 0:
                        continue
                    elif vn == 'a' and i == 0:
                        vnn = 'a'
                    else:
                        vnn = vn + str(i)
                    if vnn in out.var_names:
                        labstr_fit += pard[vn].format(out.params[vnn].value)
                    elif fix is not None and vnn in fix.keys():
                        labstr_fit += 'fix '+ pard[vn].format(fix[vnn])
                    else:
                        labstr_fit += pard[vn].format(0)
        else:
            labstr_fit = None

        if 'legd' in doplot:
            if qv is not None:
                labstr_data = r'$\mathsf{{q}} = {:.3f}\,\mathsf{{nm}}^{{-1}}$'.format(qv)
            elif data_label is not None:
                labstr_data = data_label
            else:
                labstr_data = 'data'
        elif 'legid' in doplot:
            if cf_id is not None:
                labstr_data = r'id: {}'.format(cf_id)
            else:
                labstr_data = None
        else:
            labstr_data = None

        if 'fit' in doplot and dofit:
            if 'g1' in doplot:
                g2f = np.sqrt((g2f-out.params['a'].value)/out.params['b0'].value)
            pl.append(ax.plot(xf, g2f, '-', label=labstr_fit, linewidth=1))

        if 'data' in doplot:
            if 'g1' in doplot:
                cf = np.sqrt((cf-out.params['a'].value)/out.params['b0'].value)
            if err is not None:
                pl.append(ax.errorbar(t, cf, yerr=err, linestyle='', marker=marker,
                                      label=labstr_data, alpha=alpha, markersize=markersize))
            else:
                pl.append(ax.plot(t, cf, marker, label=labstr_data, alpha=alpha,
                                  markersize=markersize))
                
        if color is None:
            if h_plot is not None:
                color = h_plot.get_color()
            elif 'data' in doplot:
                color = pl[0][0].get_color()
            else:
                color = 'gray'
                
        for p in pl:
            if type(p) == list:
                p[0].set_color(color)
            elif type(p) == matplotlib.container.ErrorbarContainer:
                    p[0].set_color(color)
                    p[2][0].set_color(color)

        if 'g1' in doplot:
            ax.set_xscale('linear')
            ax.set_yscale('log')
        else:
            ax.set_xscale('log')
            ax.set_yscale('linear')

        if 'leg' in doplot:
            ax.legend()

        if 'report' in doplot and dofit:
            print(lmfit.fit_report(out))
        
        if 'redchi' in doplot and dofit:
            print(out.redchi)

        ax.set_xlabel(r'delay time $\tau$ [s]')
        ax.set_ylabel(r'$g_2(\tau)$')

        niceplot(ax, autoscale=0)
        ax.set_xlim(*xl)
        if yl is not None:
            ax.set_ylim(*yl)
        
    pars_arr =  np.zeros((3*modes+2,2))
    if dofit:
        for i,vn in enumerate(pars.keys()):
            pars_arr[i,0] = out.params[vn].value
            try:
                pars_arr[i,1] = 1.*out.params[vn].stderr
            except TypeError:
                pars_arr[i,1] = 1
        gof = np.array([out.chisqr, out.redchi, out.bic, out.aic])
        return pars_arr, gof, out, lmfit.fit_report(out), pl
    else:
        return 0, 0, 0, 0, pl



#! /usr/bin/env python
import numpy as np
import lmfit
from .PoissonGammaDistribution import PoissonGamma as poisgam
from ..Xplot.PlotPoissonGamma import plot_poissongamma


def fit_pg(prob, err=None, ind_var='kb', krange=None, init={}, fix=None, logscale=False, qv=None,
            modes=1, pbthres=0, vary_par=False, method='leastsq', doplot=False):
    """ Fit the Poisson-Gamma distribution
    by minimizing the residuals"""

    if krange is None:
        krange = np.arange(prob.shape[0]-2)

    kv = prob[krange+2,0]
    kb = prob[1]
    prob = prob[krange+2]

    if ind_var == 'kb':
        x = kb
        p = kv
        pn = 'k'
    elif ind_var == 'k':
        x = kv
        p = kb
        pn = 'kb'
        
    #make initial guess for parameters
    for vn in ['M', 'kb']:
        if vn not in init.keys():
            if vn == 'M':
                init[vn] = (1, 0, None)
            elif vn == 'kb':
                init[vn] = (kb.mean(),0,None)

    # initialize fit parameters
    pars = lmfit.Parameters()
    for i in range(p.size):
        pars.add('M{}'.format(i), value=init['M'][0], min=init['M'][1], max=init['M'][2])
        if i>0:
            pars['M{}'.format(i)].expr = 'M0'
        pars.add(pn+'{}'.format(i), value=p[i], vary=False)
                             
    if fix is not None:
        for vn in fix.keys():
            pars[vn].set(value=fix[vn], vary=0)

    if err is not None:
        err = np.abs(err)
        wgt = err.copy()
        wgt[wgt>0] = 1./wgt[wgt>0]**2
    else:
        wgt = None
        
    # independent variable is <k>
    def residuals(pars, x, data=None, eps=None):
        """2D Residual function to minimize
        """
        v = pars.valuesdict()
        
        residuals = np.zeros((p.size,x.size))
        
        for i,pi in enumerate(p):
            if logscale:
                model = np.log10(poisgam(x, v['M{}'.format(i)], pi, ind_var))
            elif not logscale:
                model = poisgam(x, v['M{}'.format(i)], pi, ind_var)
            if eps is not None:
                residuals[i] = np.abs(data[i]-model) * np.abs(eps[i])
            else:
                residuals[i] = np.abs(data[i]-model)
        return residuals.flatten()
    
    out = lmfit.minimize(residuals, pars, args=(x,), kws={'data':prob, 'eps':wgt}, method=method,
                         nan_policy='omit')

    pars_arr =  np.zeros((1,2))
    for i,vn in enumerate(['M0']):
        pars_arr[i,0] = out.params[vn].value
        pars_arr[i,1] = 1.*out.params[vn].stderr
    gof = np.array([out.chisqr, out.redchi, out.bic, out.aic])

    if doplot:
        plot_poissongamma(x, prob, p, pars_arr, ind_var=ind_var)

    return pars_arr, gof, out, lmfit.fit_report(out)




#! /usr/bin/env python
import numpy as np
import lmfit
from Xfit.PoissonGammaDistribution import PoissonGamma as poisgam
from Xfit.FitPoissonGammaLikelihood import fit_pg_likelihood

def rescale(y, rng=(0,1)):
    return ((rng[1]-rng[0])*(y - min(y)))/(max(y)-min(y)) + rng[0]

def correct_probability(prob, kappa):
    """Correct probability
    """
    p = prob[2:]
    c = p[2] * kappa
    p[2] -= c
    p[1] += 2*c
    p[0] -= c
    return prob

def detector_correction2(prob, M_ref, npix, err=None, kv=None, init={}, fix=None, method='Nelder-mead',
                       mspacing=200):
    """ Fit the Poisson-Gamma distribution using the likelihood ratio approach.
    """
    if kv is None:
        kv = np.arange(prob.shape[0]-2)

    #make initial guess for parameters
    for vn in ['kappa']:
        if vn not in init.keys():
            if vn == 'kappa':
                init[vn] = (0.5, 0, 1)

    # initialize fit parameters
    pars = lmfit.Parameters()
    pars.add('kappa', value=init['kappa'][0], min=init['kappa'][1], max=init['kappa'][2])
                             
    if fix is not None:
        for vn in fix.keys():
            pars[vn].set(value=fix[vn], vary=0)

    if err is not None:
        err = np.abs(err)
        wgt = err.copy()
        wgt[wgt>0] = 1./wgt[wgt>0]**2
    else:
        wgt = None
        
    M = np.logspace( 0, 2, mspacing )  
        
    def chi2(prob):
        """Calculate likelihood ratio
        """
        kb = prob[1]
        prob = prob[kv+2]
        chi2 = np.zeros(( mspacing ))
        for j,m in enumerate(M):
            for i in range(kv.size):
                probi = prob[i]
                ind = np.where(probi)
                pg = poisgam(kb[ind], m, kv[i], ind_var='kb')
                chi2[j] += np.sum(probi[ind] * np.log(pg/probi[ind]))
            chi2[j] *= -2 * npix
        return rescale(chi2, (0,1))
    
    chi2_ref = chi2(prob_ref)

    print('test')
    def residual(pars, prob, eps=None):
        """Residual function to minimize
        """
        prob = prob.copy()
        v = pars.valuesdict()
        
        prob = correct_probability(prob, v['kappa'])
        res = fit_pg_likelihood(prob, npix, init={'M':(4,1,100)}, kv=kv)
        return np.abs(M_ref-res[0][0,0])

    out = lmfit.minimize(residual, pars, args=(prob,), kws={'eps':wgt}, method=method,
                         nan_policy='omit')

    pars_arr =  np.zeros((1,2))
    for i,vn in enumerate(['kappa']):
        pars_arr[i,0] = out.params[vn].value
        # pars_arr[i,1] = pars_arr[i,0]**2*out.params[vn].stderr
    gof = np.array([out.chisqr, out.redchi, out.bic, out.aic])

    return pars_arr, gof, out, lmfit.fit_report(out)



from matplotlib import pyplot as plt
import matplotlib as mpl
import numpy as np

import matplotlib.path as path
from matplotlib.colors import ListedColormap

from time import sleep
import ipywidgets as widgets

def masker(data, mask1=[]):

    if not isinstance(mask1, np.ndarray):
        mask1 = np.ones_like(data)

    xs = []
    ys = []
    mask = []

    def invert_mask(*args):
        ind = mask1 > 0
        mask1[ind] = 0
        mask1[~ind] = 1
        mask.set_data(mask1)

    def reset_mask(*args):
        mask1[:] = 1
        mask.set_data(mask1)

    def save_mask(*args):
        savname = savwid.value
        savname = savname.replace('.npy', "") + '.npy'
        np.save(savname, mask1.astype(bool))

    def cancel_masking(*args):
        del xs[:]
        del ys[:]
        update_plot()

    def mask_roi(*args):
        lx,ly=np.shape(mask1)
        x, y = np.meshgrid(np.arange(lx), np.arange(ly))
        x, y = x.flatten(), y.flatten()
        points = np.vstack((x,y)).T
        xy = np.vstack((np.array(ys),np.array(xs))).T
        new_submask = ~path.Path(xy).contains_points(points).reshape(ly,lx).T 
        np.multiply(mask1, new_submask, out=mask1)
        del xs[:]
        del ys[:]
        update_plot()

    def mask_outside(*args):
        lx,ly=np.shape(mask1)
        x, y = np.meshgrid(np.arange(lx), np.arange(ly))
        x, y = x.flatten(), y.flatten()
        points = np.vstack((x,y)).T
        xy = np.vstack((np.array(ys),np.array(xs))).T
        new_submask = path.Path(xy).contains_points(points).reshape(ly,lx).T 
        np.multiply(mask1, new_submask, out=mask1)
        del xs[:]
        del ys[:]
        update_plot()

    def unmask_roi(*args):
        lx,ly=np.shape(mask1)
        x, y = np.meshgrid(np.arange(lx), np.arange(ly))
        x, y = x.flatten(), y.flatten()
        points = np.vstack((x,y)).T
        xy = np.vstack((np.array(ys),np.array(xs))).T
        new_submask = path.Path(xy).contains_points(points).reshape(ly,lx).T 
        mask1[new_submask] = True
        del xs[:]
        del ys[:]
        update_plot()

    def above_masking(*args):
        thres = int(mvwid.value)
        mask1[data>=thres]=False
        update_plot()

    def below_masking(*args):
        thres = int(mvwid.value)
        mask1[data<=thres]=False
        update_plot()

    def onkey(event,line):
        if event.key == 'c':
            cancel_masking()
        elif event.key == 'm':
            mask_roi()
        elif event.key == 'o':
            mask_outside()
        elif event.key == 'u':
            unmask_roi()
        elif event.key == 'i':
            invert_mask()
        elif event.key == 'a':
            above_masking()
        elif event.key == 'b':
            below_masking()


    def onclick(event,line):
        xs.append(event.xdata)
        ys.append(event.ydata)

        line.set_data(xs, ys)
        line.figure.canvas.draw_idle()

    def update_plot(*args):
        line.set_data(xs, ys)
        mask.set_data(mask1.astype(int))


    mvwid = widgets.Text(
        value='0',
        description='mask value:',
        placeholder='type herer',
        disabled=False
    )
    above_button = widgets.Button(
        description='mask >= (a)',
        disabled=False,
        tooltip='all values set to default',
    )
    above_button.on_click(above_masking)
    below_button = widgets.Button(
        description='mask <= (b)',
        disabled=False,
        tooltip='all values set to default',
    )
    below_button.on_click(below_masking)
    row_thres = widgets.HBox([mvwid,above_button,below_button])

    cancel_button = widgets.Button(
        description='cancel (c)',
        disabled=False,
        tooltip='cancel roi',
    )
    cancel_button.on_click(cancel_masking)
    reset_button = widgets.Button(
        description='reset',
        disabled=False,
        tooltip='all values set to default',
    )
    reset_button.on_click(reset_mask)

    row3 = widgets.HBox([cancel_button,
                         reset_button])

    mask_button = widgets.Button(
        description='mask (m)',
        disabled=False,
        tooltip='mask roi',
        icon='',
    )
    mask_button.on_click(mask_roi)
    unmask_button = widgets.Button(
        description='unmask (u)',
        disabled=False,
        tooltip='unmask roi',
    )
    unmask_button.on_click(unmask_roi)
    outside_button = widgets.Button(
        description='mask outside (o)',
        disabled=False,
        tooltip='mask everything outside the roi',
    )
    outside_button.on_click(mask_outside)
    invert_button = widgets.Button(
        description='invert (i)',
        disabled=False,
        tooltip='invert the complete mask',
        icon='',
    )
    invert_button.on_click(invert_mask)

    row2 = widgets.HBox([mask_button,
                         unmask_button,
                         outside_button,
                         invert_button])

    savwid = widgets.Text(
        value='example_mask.npy',
        placeholder='type here',
        description='Filename:',
        disabled=False
    )
    cdir_button = widgets.Button(
        description='change directory',
        disabled=False,
        tooltip='change saving directory',
    )
    save_button = widgets.Button(
        description='save mask',
        disabled=False,
        tooltip='save mask to npy file',
    )
    save_button.on_click(save_mask)
    row_save = widgets.HBox([savwid,
                         cdir_button,
                         save_button])

    out = widgets.Output()
    display(out)
    with out:

        display(widgets.VBox([row2,row3,row_thres,row_save]))

        fig, ax = plt.subplots(figsize=(8,8))
        ax.set_position([.06, .05, .94, .94])

        cmap = mpl.colors.ListedColormap(['orange', 'white'])

        # Get the colormap colors
        my_cmap = cmap(np.arange(cmap.N))

        # Set alpha
        my_cmap[:,-1] = 0.7

        # Create new colormap
        cmap = ListedColormap(my_cmap)
        cmap.set_over('orange', alpha=0)

        bounds=[0,.5,1]
        norm = mpl.colors.BoundaryNorm(bounds, cmap.N)
        ax.imshow(data, interpolation='none', norm=mpl.colors.LogNorm(), aspect='auto')
        mask = plt.imshow(mask1.astype(int), interpolation='none', norm=norm, cmap=cmap, aspect='auto')
        plt.axis('equal')
        line, = ax.plot([],[],'.-k')
        cid_m = fig.canvas.mpl_connect('button_press_event', lambda event: onclick(event,line))
        cid_k = fig.canvas.mpl_connect('key_press_event', lambda event: onkey(event,line))
        plt.show()



import numpy as np
import pandas as pd
import re
from io import StringIO


def loadmonitor(f, scan_time='3000-01-01 00:00:00', scanid=-1):
    """Load ID10 meta information from monitor file f. 
    The time stamp of the measurement is required.
    """
    scan_time = pd.to_datetime(scan_time)
    dummy_time = 0
    i = 0
    with open(f, 'r') as monitor:
        line = monitor.readline()
        while line:
            if line.startswith('#S'):
                S = re.search('(?<=S )\d*',line).group(0)  
                date = monitor.readline()
                line += date
                date = date.lstrip('#D ')
                nearest_time =  pd.to_datetime(date)
                if nearest_time > scan_time or int(S) == scanid:
                    scan = ''
                    while line:
                        scan += line
                        line = monitor.readline()
                        if len(line.strip()) == 0 :
                            print('Scan ID is ', S)
                            print('Closest time is ', nearest_time)
                            break  
                    break
            else:
                pass
            line = monitor.readline()
            
    hline, lline = np.where([0 if x.startswith('#') else 1 for x in scan.split('\n')])[0][[0,-2]]
    header = re.split(r'\s{2,}', scan.split('\n')[hline-1])
    header[0] = header[0][3:]
    df = pd.read_csv(StringIO(scan), names=header, skiprows=hline, nrows=lline-hline+1, delim_whitespace=1)

    return df, scan, (hline-1,lline+1)





import numpy as np
from matplotlib import pyplot as plt
from ..Decorators import Decorators
from ..Analist import AnaList
from .pysaxs3 import get_soq
import copy
from ..Xplot.niceplot import niceplot


class Soq(AnaList):
    """
    Saxs class: Display and analyze Small Angle X-Ray scattering singals.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __str__(self):
        return 'Saxs class for displaying saxs signal.'

    def __repr__(self):
        return self.__str__()

    def get_sec(self, img):

        if np.shape(img) != np.shape(self.Xana.setup.mask):
            qsec = self.Xana.setup['qsec'][0]
            mask = self.Xana.mask.copy()
            dim = np.shape(img)
            mask = mask[qsec[0]:dim[0]+qsec[0], qsec[1]:dim[1]+qsec[1]]

            setup = copy.deepcopy(self.Xana.setup)
            setup['ctr'] = (setup['ctr'][0]-qsec[1], setup['ctr'][1]-qsec[0])
            return mask, setup
        else:
            return self.Xana.mask, self.Xana.setup


    @Decorators.init_figure()
    @Decorators.input2list
    def plot_soq( self, series_id, shade=False, cmap='tab10', cmap_shade='inferno',
                  color_mode=0, ax=None, change_marker=0, markersize=3, legend='id',
                  qexp=None, Iscaling=None, A=1., norm=False, normto=None, bg=None,
                  logax='xy', Ae=1., show_legend=True, color='b', normto_exposure=False,
                  **kwargs ):

        if color_mode < 2:
            if color_mode == 0:
                color_multiplier, color_repeater = len(series_id), 1
            elif color_mode == 1:
                self.colors = [color]*len(series_id)
            self.update_colors(cmap, color_multiplier, color_repeater)
        elif color_mode == 2:
            self.colors = [color]*len(series_id)

        # niceplot(ax)

        ncorrections = sum([bool(normto),bool(bg)])

        if normto is not None:
            del series_id[series_id.index(normto)]
            series_id.insert(0, normto)

        Ib = 0
        if bg is not None:
            del series_id[series_id.index(bg)]
            series_id.insert(bg,0)

        output = []
        for i, sid in enumerate(series_id):
            saxsd = self.Xana.get_item(sid)

            if 'soq' not in saxsd:
                Isaxs = saxsd['Isaxs']

                if Isaxs.ndim == 3:
                    Isaxs = self.Xana.arrange_tiles(Isaxs)

#                mask, setup = self.get_sec(Isaxs)
                q, I, e  = get_soq(Isaxs, self.Xana.setup)
            else:
                tmp = saxsd['soq']
                q = tmp[:,0]
                I = tmp[:,1]
                e = tmp[:,2]

            legstr = None
            if legend == 'id':
                legstr = 'id {}'.format(int(sid))
            elif legend == 't_exposure':
                legstr = r'$t_{{e}} = {:.2e}s$'.format(self.Xana.db.loc[sid, 't_exposure'])
            elif legend == 'sample':
                legstr = self.Xana.db.loc[sid, 'sample']

            ylabel = r'$I(q)$'
            if Iscaling != None:
                ylabel += r" $\cdot q^{{{}}}$".format(Iscaling)
                I *= q**Iscaling
                e *= q**Iscaling

            xlabel = r'$q$ ($\mathrm{nm}^{-1}$)'
            if qexp != None:
                xlabel = r'$q^{{{}}}$ ($\mathrm{{nm}}^{{-{}}}$)'.format(qexp, qexp)
                q *= q**(qexp-1)

            if bg is not None:
                if i == 0:
                    Ib = I
                    ncorrections = 0
                    continue
                else:
                    I -= Ib

            if normto_exposure:
                I /= self.Xana.db.loc[sid, 't_exposure']

            if norm:
                e /= I.mean()
                I /= I.mean()

            if normto is not None:
                if (i==0):
                    normto = (q,I,e)
                    continue
                else:
                    I /= normto[1]
                    e /= normto[1]

            ax.errorbar(q, I*A, e*A*Ae, fmt='-', marker=self.markers[i%len(self.markers)],
                      color=self.colors[i], label=legstr, markersize=markersize)
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)
            if 'x' in logax:
                ax.set_xscale('log')
            if 'y' in logax:
                ax.set_yscale('log')

            output.append((q,I*A,e*A))

        if show_legend:
            ax.legend(bbox_to_anchor=(1.04,1), borderaxespad=0)
            plt.tight_layout(rect=[0,0,0.98,1])
        if shade:
            ax = plt.gca()
            shadeqrois(ax, self.Xana.setup['qv'], self.Xana.setup['dqv'], cmap=cmap_shade)



import h5py
import numpy as np
import os
import re
from os.path import isfile
#from EdfMethods import loadedf, headeredf


#######################################
#--- get information on hdf5 series ---#
#######################################    

# --- older version of gertfiles; just kept here in case needed for p10 experiment ---
# def getfiles(datdir, suffix, numfmt='(_\d)*'):
#     filelist = [ os.path.abspath(datdir + item) for item in os.listdir(datdir)
#                if os.path.isfile(os.path.join(datdir, item)) and item.endswith(suffix) ]
#     filelist = sorted(filelist, key=lambda x: int(''.join(re.findall( numfmt ,x))))
#     return filelist

def getfiles(datdir, suffix, numfmt='(_\d)*'):
    check_suffix = re.compile(suffix)
    filelist = [ os.path.abspath(datdir + item) for item in os.listdir(datdir)
                 if os.path.isfile(os.path.join(datdir, item)) and bool(check_suffix.search(item))]
    filelist = sorted(filelist, key=lambda x: int(''.join(re.search( numfmt ,x).group(0).split('_'))))
    return filelist

def files2series(filelist, masters, seriesfmt,):
    series = []
    series_id = np.empty(len(masters), dtype=np.int32)
    find_seriesid = re.compile(seriesfmt)
    for i,m in enumerate(masters):
        idstr = find_seriesid.search(m).group()
        series_id[i] = int(idstr)
        nblocks = len(re.findall('(_\d{4,})', m))
        searchstr = idstr + '.*(_\d{{4,}}){{{}}}'.format(nblocks-1)
        series.append([x for x in filelist if re.search(searchstr, x) is not None])
    return series, series_id

def get_attr_from_dict(obj, meta):
            
    def init_meta(p):
        for key, value in p.items():
            meta[key] = np.empty(nfiles, dtype=np.dtype(value[1]))
            
    def get_attr(header, p):
        try:
            attr = p[1](header[p[0]])
        except:
            attr = 0
        return attr

    nfiles = len(meta['master'])
    p = obj.attributes
    init_meta(p)
    for i, m in enumerate(meta.copy()['master']):
        filename = obj.datdir + m
        header = headeredf(filename)
        for key, value in p.items():
            meta[key][i] = get_attr(header, value)
                
def get_attributes(filename, c, obj):
    p = obj.attributes[c]
    f = headeredf(filename)
    try:
        attr = p[1](f[p[0]])
    except:
        attr = 0
    return attr

def get_header_h5(filename):
    # f = h5py.File(filename, 'r', driver='stdio')
    # header = dict(f['/entry/instrument/detector/'])
    # f.close()
    return 0

def get_attributes_h5(obj, meta):
            
    def init_meta(p):
        for key, value in p.items():
            meta[key] = np.empty(nfiles, dtype=np.dtype(value[1]))
            
    def get_attr(f, p):
        if p[0] == obj.h5opt['data']:
            attr = f[p[0]].attrs[c]
        else:
            if p[0]:
                attr = f[p[0]]
                if attr.dtype.kind == 'V':
                    attr = np.mean(attr.value['value']).astype(np.dtype(p[1]))
                elif attr.dtype.kind == 'f':
                    attr = attr[()]#.astype(np.dtype(p[1]))
                elif attr.dtype.kind == 'u':
                    attr = attr[()]#.astype(np.dtype(p[1]))
                elif attr.dtype.kind == 'i':
                    attr = attr[()]
            else:
                attr = f[obj.h5opt['data']].shape[0]
        return attr

    nfiles = len(meta['master'])
    p = obj.attributes
    init_meta(p)
    for i, m in enumerate(meta.copy()['master']):
        filename = obj.datdir + m
        with h5py.File(filename, 'r', driver=obj.h5opt['driver']) as f:
            for key, value in p.items():
                meta[key][i] = get_attr(f, value)


def common_mode_from_hist(im, searchoffset=50):                        
    bins = np.arange(-searchoffset,searchoffset) 
    hist, _ = np.histogram(im,bins)
    CM = np.where(hist==hist.max())[0][0] - searchoffset
    return CM # has to be subtracted from the tail



import numpy as np
import emcee
import corner
import scipy.optimize as op


def mcmc_sl(x, y, dy, doplot=0):
    A = np.vstack((np.ones_like(x), x)).T
    C = np.diag(dy * dy)
    cov = np.linalg.inv(np.dot(A.T, np.linalg.solve(C, A)))
    b_ls, m_ls = np.dot(cov, np.dot(A.T, np.linalg.solve(C, y)))
    
    def lnlike(theta, x, y, yerr):
        m, b, lnf = theta
        model = m * x + b
        inv_sigma2 = 1.0/(yerr**2 + model**2*np.exp(2*lnf))
        return -0.5*(np.sum((y-model)**2*inv_sigma2 - np.log(inv_sigma2)))
    
    nll = lambda *args: -lnlike(*args)
    result = op.minimize(nll, [m_ls, b_ls, np.log(.5)], args=(x, y, dy))
    m_ml, b_ml, lnf_ml = result["x"]
    
    def lnprior(theta):
        m, b, lnf = theta
        if 0. < m and -50. < b < 50.0 and -15.0 < lnf < 2.0:
            return 0.0
        return -np.inf
    
    def lnprob(theta, x, y, yerr):
        lp = lnprior(theta)
        if not np.isfinite(lp):
            return -np.inf
        return lp + lnlike(theta, x, y, yerr)
    
    ndim, nwalkers = 3, 100
    pos = [result["x"] + 1e-4*np.random.randn(ndim) for i in range(nwalkers)]
    
    sampler = emcee.EnsembleSampler(nwalkers, ndim, lnprob, args=(x, y, dy))
    sampler.run_mcmc(pos, 500)
    samples = sampler.chain[:, 50:, :].reshape((-1, ndim))
    
    if doplot:
        fig = corner.corner(samples, labels=["$m$", "$b$", "$\ln\,f$"],
                          truths=[m_ml, b_ml, lnf_ml])

        xl = np.linspace(x.min(),x.max(),100)
        for m, b, lnf in samples[np.random.randint(len(samples), size=100)]:
            doplot.plot(xl, m*xl+b, color="k", alpha=0.1)
        #plt.plot(xl, m_true*xl+b_true, color="r", lw=2, alpha=0.8)
        #plt.errorbar(x, y, yerr=yerr, fmt=".k")

    samples[:, 2] = np.exp(samples[:, 2])
    m_mcmc, b_mcmc, f_mcmc = map(lambda v: (v[1], v[2]-v[1], v[1]-v[0]),
                             zip(*np.percentile(samples, [16, 50, 84],
                                                axis=0)))
    return m_mcmc, b_mcmc, f_mcmc, m_ls, b_ls



# toolbox.py
import numpy as np
import matplotlib
from matplotlib import pyplot as plt

def gen_log_space(limit, n):
    result = [1]
    if n>1:  # just a check to avoid ZeroDivisionError
        ratio = (float(limit)/result[-1]) ** (1.0/(n-len(result)))
    while len(result)<n:
        next_value = result[-1]*ratio
        if next_value - result[-1] >= 1:
            # safe zone. next_value will be a different integer
            result.append(next_value)
        else:
            # problem! same integer. we need to find next_value by artificially incrementing previous value
            result.append(result[-1]+1)
            # recalculate the ratio so that the remaining values will scale correctly
            ratio = (float(limit)/result[-1]) ** (1.0/(n-len(result)))
    # round, re-adjust to 0 indexing (i.e. minus 1) and return np.uint64 array
    return np.array(list(map(lambda x: round(x)-1, result)), dtype=np.uint64)


def add_colorbar(ax, vec, label=None, cmap='magma', discrete=False, tick_indices=None, shrink=1., fraction=.15):
    if tick_indices is None:
        tick_indices = np.arange(len(vec))
    vec = np.array(vec)
    # divider = make_axes_locatable(ax)
    # cax = divider.append_axes('right', size='5%', pad=0.05)

    change_ticks = False
    if discrete:
        ticks = vec
        norm = mpl.colors.nonorm()
        change_ticks = true
        if isinstance(cmap, str):
            cmap = plt.get_cmap(cmap, len(vec))
        elif isinstance(cmap, (list, np.ndarray)):
            cmap = mpl.colors.ListedColormap(cmap)
    else:
        if isinstance(cmap, str):
            cmap = plt.get_cmap(cmap)
            ticks = none
        elif isinstance(cmap, (list, np.ndarray)):
            cmap = mpl.colors.ListedColormap(cmap)
        norm = mpl.colors.Normalize(vmin=vec.min(), vmax=vec.max())

    cax = mpl.colorbar.make_axes(ax, shrink=shrink, fraction=fraction)[0]
    cb1 = mpl.colorbar.ColorbarBase(cax, norm=norm, cmap=cmap,
                                    orientation='vertical',)

    cb1.set_label(label)
    if change_ticks:
        cb1.set_ticks(tick_indices)
        cb1.set_ticklabels(list(map(lambda x: "%.3f" % x, vec[tick_indices])))
    cb1.ax.invert_yaxis()
    cb1.ax.set_in_layout(True)
    # cax.yaxis.label.set_in_layout(true)
